from __future__ import annotations
from abc import abstractmethod
import os
import logging

from pathlib import Path
from typing import Dict, List, Optional, Any, Literal, Union
from typing_extensions import Annotated
from mmm.BaseModel import BaseModel
from pydantic import Field

import numpy as np
import torch
import torch.optim as optim
from torch.optim.lr_scheduler import (
    CosineAnnealingLR,
    LambdaLR,
    ExponentialLR,
    ReduceLROnPlateau,
)
from torch.distributed.optim import ZeroRedundancyOptimizer
import torch.distributed as dist

from mmm.mtl_modules.shared_blocks.SharedBlock import SharedBlock
from mmm.mtl_modules.shared_blocks.SharedModules import SharedModules
from mmm.mtl_modules.tasks.MTLTask import MTLTask


TorchLRScheduler = Union[optim.lr_scheduler._LRScheduler, ReduceLROnPlateau]


class SchedulerConfig(BaseModel):
    @abstractmethod
    def build_instance(self, optim: optim.Optimizer, max_epochs: int) -> TorchLRScheduler:
        raise NotImplementedError


class PolySchedulerConfig(SchedulerConfig):
    scheduler_type: Literal["poly"] = "poly"
    exponent: float = 0.9
    last_epoch: int = -1

    def build_instance(self, optim: optim.Optimizer, max_epochs: int) -> optim.lr_scheduler._LRScheduler:
        initial_lrs = [p["lr"] for p in optim.param_groups]

        def poly_lr(epoch, initial_lr):
            return initial_lr * (1 - epoch / max_epochs + 1) ** self.exponent

        return LambdaLR(
            optim,
            lr_lambda=[lambda e: poly_lr(e, lr) for lr in initial_lrs],
            last_epoch=self.last_epoch,
        )


class ExponentialLRConfig(SchedulerConfig):
    scheduler_type: Literal["exp"] = "exp"
    gamma: float = 0.9

    def build_instance(self, optim: optim.Optimizer, max_epochs: int) -> optim.lr_scheduler._LRScheduler:
        return ExponentialLR(optim, gamma=self.gamma, last_epoch=-1)


class ReduceLROnPlateauConfig(SchedulerConfig):
    scheduler_type: Literal["reduceonplateau"] = "reduceonplateau"
    factor: float = 0.1
    patience: int = 4
    threshold: float = 1e-4
    cooldown: int = 0
    min_lr: float = 1e-7
    eps: float = 1e-8

    def build_instance(self, optim: optim.Optimizer, max_epochs: int) -> ReduceLROnPlateau:
        return ReduceLROnPlateau(
            optim,
            mode="min",
            factor=self.factor,
            patience=self.patience,
            threshold=self.threshold,
            cooldown=self.cooldown,
            min_lr=self.min_lr,
            eps=self.eps,
        )


class DecaySchedulerConfig(SchedulerConfig):
    """
    Config of:
    https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.StepLR.html#torch.optim.lr_scheduler.StepLR
    """

    scheduler_type: Literal["step"] = "step"
    factor: float = 0.98
    min_lr: float = 1e-6
    last_epoch: int = -1

    def build_instance(self, optim: optim.Optimizer, max_epochs: int) -> optim.lr_scheduler._LRScheduler:
        def f(epoch):
            r = self.factor**epoch
            return r if r > self.min_lr else self.min_lr

        return LambdaLR(optim, lr_lambda=f, last_epoch=self.last_epoch)


class CosineAnnealingLRSchedulerConfig(SchedulerConfig):
    """
    Config of
    OptimizerType = Union[OptimizerAdamConfig, OptimizerAdamWConfig, OptimizerSGDConfig]
    """

    scheduler_type: Literal["cosine_annealing"] = "cosine_annealing"
    eta_min: float = 1e-6
    last_epoch: int = -1
    cycle_size: int = 5

    def build_instance(self, optim: optim.Optimizer, max_epochs: int) -> optim.lr_scheduler._LRScheduler:
        return CosineAnnealingLR(
            optim,
            T_max=self.cycle_size,
            eta_min=self.eta_min,
            last_epoch=self.last_epoch,
        )


SchedulerType = Union[
    PolySchedulerConfig,
    ExponentialLRConfig,
    ReduceLROnPlateauConfig,
    DecaySchedulerConfig,
    CosineAnnealingLRSchedulerConfig,
]
SchedulerList = List[SchedulerType]


class OptimizerConfig(BaseModel):
    lr: float = 1e-4

    def get_type(self):
        raise NotImplementedError

    def get_params(self):
        raise NotImplementedError


class OptimizerAdamWConfig(OptimizerConfig):
    optim_type: Literal["adamw"] = "adamw"
    weight_decay: float = 1e-2

    def get_type(self):
        return optim.AdamW

    def get_params(self):
        return dict(lr=self.lr, weight_decay=self.weight_decay)


class OptimizerSGDConfig(OptimizerConfig):
    optim_type: Literal["sgd"] = "sgd"
    momentum: float = 0.8
    weight_decay: float = 0.0
    dampening: float = 0.0
    nesterov: bool = True
    lr: float = 2.5e-3

    def get_type(self):
        return optim.SGD

    def get_params(self):
        return dict(
            lr=self.lr,
            momentum=self.momentum,
            weight_decay=self.weight_decay,
            dampening=self.dampening,
            nesterov=self.nesterov,
        )

    def build_instance(self, params) -> optim.Optimizer:
        return optim.SGD(
            params,
            lr=self.lr,
            momentum=self.momentum,
            weight_decay=self.weight_decay,
            dampening=self.dampening,
            nesterov=self.nesterov,
        )


OptimizerType = Union[OptimizerAdamWConfig, OptimizerSGDConfig]


class MTLOptimizer:
    class Config(BaseModel):
        optim_config: Annotated[OptimizerType, Field(discriminator="optim_type")] = OptimizerAdamWConfig()
        lr_scheduler_configs: List[Annotated[SchedulerType, Field(discriminator="scheduler_type")]] = []
        lr_factor_tasks: float = Field(
            default=1.0,
            description="Scales the learning rate of the task optimizers relative to the global learning rate.",
        )

    def __init__(
        self,
        shared_blocks: SharedModules,
        args: MTLOptimizer.Config,
        max_epochs: int,
        gradient_scale_factor: float = 1.0,
    ) -> None:
        self.args = args
        self.gradient_scale_factor = gradient_scale_factor
        msg = "Setting up Optimizer for: "

        # Parameter are a list of dictionaries, where each dictionary correspond to an MTLModule
        # Each dictionary can contain extra info such as a custom learning rate for that MTLModule
        ps: List[Dict[str, Any]] = []
        for k, shared_block in shared_blocks.items():
            ps.append(
                {
                    "initial_lr": self.args.optim_config.lr,
                    "params": list(shared_block.parameters()),
                }
            )
            msg = f"{msg} block:{k}"

        if dist.is_initialized():
            self.optimizer: optim.Optimizer = ZeroRedundancyOptimizer(
                ps,
                optimizer_class=self.args.optim_config.get_type(),
                **self.args.optim_config.get_params(),
            )
        else:
            self.optimizer: optim.Optimizer = self.args.optim_config.get_type()(
                ps, **self.args.optim_config.get_params()
            )

        self.schedulers = [c.build_instance(self.optimizer, max_epochs) for c in self.args.lr_scheduler_configs]
        self.task_optims: dict[str, optim.Optimizer] = {}
        # A list which contains the names of the tasks which still have UNUSED gradients in the current step
        self.still_has_gradient: list[str] = []
        logging.info(msg)

    def add_task(self, task: MTLTask) -> None:
        """
        Adds a new task's optimizable parameters. Cannot check if the task already exists.
        """
        assert task.get_name() not in list(self.task_optims.keys()), f"{self} already knows {task.get_name()}"
        task_optim_config = self.args.optim_config.model_copy()
        task_optim_config.lr = task_optim_config.lr * self.args.lr_factor_tasks
        task_optim = self.args.optim_config.get_type()(task.parameters(), **task_optim_config.get_params())
        self.task_optims[task.get_name()] = task_optim

    def reset_shared_grads(self):
        # We do not want the memory to accumulate with the number of tasks in a step -> set_to_none=True
        self.optimizer.zero_grad(set_to_none=True)

    def after_backward(self, of_task: MTLTask, reduce_task_immediately: bool):
        if reduce_task_immediately:
            self.task_optims[of_task.get_name()].step()
            # The docs mention that in that case steps might be skipped if gradients are set to none, beware!
            # In a multi-task setting, not all shared blocks might be used in every step resulting in skipped steps.
            self.task_optims[of_task.get_name()].zero_grad(set_to_none=True)
        else:
            if of_task.get_name() not in self.still_has_gradient:
                self.still_has_gradient.append(of_task.get_name())

    def after_iteration_step(self):
        # Gradients are averaged across all processes or at least nodes by DDP
        # We want to compensate for that by multiplying all shared parameter's gradients with the global_world_size
        # This only needs to be applied to the shared parameters, as the parameters of the tasks are optimized locally.
        if self.gradient_scale_factor != 1.0:
            for param_group in self.optimizer.param_groups:
                for param_index, p in enumerate(param_group["params"]):
                    if p.grad is not None:
                        # logging.debug(f"Gradient norm of {param_index}: {torch.norm(p.grad)}")
                        p.grad *= self.gradient_scale_factor
                        # logging.debug(f"Gradient norm of {param_index}: {torch.norm(p.grad)}")
        else:
            assert (
                not dist.is_initialized()
            ), "Scale factor is 1.0 while using DDP. Gradient scaling should be done with DDP."

        self.optimizer.step()

        for task_with_unused_grad in self.still_has_gradient:
            # logging.info(f"Task {task_with_unused_grad} still has unused gradients. Performing step.")
            self.task_optims[task_with_unused_grad].step()
            self.task_optims[task_with_unused_grad].zero_grad(set_to_none=True)
        self.still_has_gradient = []

    def after_loop_step(
        self,
        epoch: int,
        train_losses: List[List[float]],
        val_losses: Optional[List[List[float]]],
        verbose=False,
    ):
        """
        If a scheduler depends on the loss, then a list with the list of losses for all tasks needs to be provided.
        """
        for scheduler in self.schedulers:
            if isinstance(scheduler, ReduceLROnPlateau):
                losses = val_losses if val_losses else train_losses
                task_averages = [np.average(ll) for ll in losses]
                # Each task contributes equally to the avg_loss
                avg_loss = np.average(task_averages)
                scheduler.step(avg_loss)
            else:
                scheduler.step()
        if verbose:
            lrs = [param_group["lr"] for param_group in self.optimizer.param_groups]
            logging.debug(f"Learning rates for param_groups: {lrs}")

    def get_current_logstate(self) -> Dict[str, Any]:
        res = {}
        # if self.schedulers:
        #     scheduler_info = {f"schedulerlastlr{i}": v
        #                       for i, v in enumerate(self.schedulers[-1].get_last_lr())}
        #     res.update(scheduler_info)
        for i, param_group in enumerate(self.optimizer.param_groups):
            res[f"lr{i}"] = param_group["lr"]
        return res

    def create_checkpoint(self, folder_path: Path, rank: int) -> None:
        if dist.is_initialized():
            self.optimizer.consolidate_state_dict()
        if rank == 0:
            if not folder_path.exists():
                os.mkdir(folder_path)
            torch.save(self.optimizer.state_dict(), folder_path / "optim.ckpt")
            if self.schedulers:
                assert self.schedulers is not None
                torch.save(
                    [s.state_dict() for s in self.schedulers],
                    folder_path / "schedulers.ckpt",
                )

        if dist.is_initialized():
            dist.barrier()
        # Save checkpoints of the tasks
        for task_name, task_optim in self.task_optims.items():
            logging.debug(f"Saving optimizer checkpoint for task {task_name}")
            torch.save(task_optim.state_dict(), folder_path / f"optim_{task_name}.ckpt")

    def load_checkpoint(self, folder_path: Path) -> None:
        for task_name, task_optim in self.task_optims.items():
            logging.debug(f"Loading optimizer checkpoint for task {task_name}")
            task_optim.load_state_dict(torch.load(folder_path / f"optim_{task_name}.ckpt"))
        self.optimizer.load_state_dict(torch.load(folder_path / "optim.ckpt"))
        if self.schedulers:
            schedulers_path = folder_path / "schedulers.ckpt"
            if schedulers_path.exists():
                try:
                    states = torch.load(schedulers_path)
                    for scheduler, state in zip(self.schedulers, states):
                        scheduler.load_state_dict(state)
                except:
                    logging.warn(f"Checkpoint {schedulers_path} not valid for LR schedulers: {self.schedulers}")
            else:
                logging.warn(f"No state found for LR schedulers: {self.schedulers}")



"""
You can use the common S3 defaults by using the environment variables: S3URL, AWS_ACCESS_KEY_ID, and AWS_SECRET_ACCESS_KEY.
Use this code:

```python
client: Minio = Minio(*get_args(), **get_kwargs())
```
"""

from __future__ import annotations
from typing import BinaryIO, Callable
import os
from pathlib import Path
from pydantic import Field
from mmm.utils import disk_cacher
from mmm.BaseModel import BaseModel
from io import BytesIO
from PIL import Image

try:
    from minio import Minio
    from minio.error import S3Error
except ImportError:
    pass

from torchvision.datasets.folder import make_dataset
from torchvision.datasets import ImageFolder


@disk_cacher(cache_path="shared")
def make_dataset_cached(*args, **kwargs):
    return make_dataset(*args, **kwargs)


class S3ImageFolder(ImageFolder):
    """
    Listing many files with S3 is slow. This class behaves like ImageFolder, but it caches the file list.
    """

    @staticmethod
    def make_dataset(
        directory: str,
        class_to_idx: dict[str, int],
        extensions: tuple[str, ...] | None = None,
        is_valid_file: Callable[[str], bool] | None = None,
    ) -> list[tuple[str, int]]:
        if class_to_idx is None:
            # prevent potential bug since make_dataset() would use the class_to_idx logic of the
            # find_classes() function, instead of using that of the find_classes() method, which
            # is potentially overridden and thus could have a different logic.
            raise ValueError("The class_to_idx parameter cannot be None.")
        return make_dataset_cached(directory, class_to_idx, extensions=extensions, is_valid_file=is_valid_file)


def get_args():
    return (
        os.getenv("S3URL", default="http://localhost:9000").replace("http://", "").replace("https://", ""),
        os.getenv("AWS_ACCESS_KEY_ID", default="minioadmin"),
        os.getenv("AWS_SECRET_ACCESS_KEY", default="minioadmin"),
    )


def get_kwargs():
    return {
        "secure": os.getenv("S3URL", default="http://localhost:9000").startswith("https://"),
    }


def download_object(client: Minio, bucket_name: str, object_name: str) -> BytesIO:
    """
    Downloads an object from the bucket without saving it to disk.
    """
    file_obj = BytesIO()
    try:
        response = client.get_object(
            bucket_name,
            object_name,
        )

        for d in response.stream(amt=1024 * 1024):
            file_obj.write(d)

        file_obj.seek(0)
        return file_obj

    finally:
        if response is not None:
            response.close()
            response.release_conn()


@disk_cacher(cache_path="shared")
def index_files(bucket_name: str, p: Path, recursive=True) -> list:
    """
    Indexes files in a bucket.
    """
    client = Minio(*get_args(), **get_kwargs())
    # The prefix consists of the path except for the first part, which needs to be the bucket name.
    prefix = "/".join(p.parts[2 if p.is_absolute else 1 :])

    bucket_path = Path("".join(p.parts[: 2 if p.is_absolute else 1]))
    return [
        bucket_path / object.object_name
        for object in client.list_objects(bucket_name, prefix=prefix, recursive=recursive)
    ]


@disk_cacher(cache_path="shared")
def cacheglob(path: Path, pattern: str) -> list:
    return list(path.glob(pattern))


def upload_img(
    mclient: Minio,
    bucket: str,
    prefix,
    img: Image,
    img_name,
    format: str = "PNG",
    base_url: str = "http://s3.datanodefec:9500",
    data_key: str = "image",
):
    # Convert image to bytes
    img_bytes = BytesIO()
    img.save(img_bytes, format=format)
    img_bytes.seek(0)
    r = mclient.put_object(bucket, f"{prefix}/{img_name}.png", img_bytes, length=img_bytes.getbuffer().nbytes)
    return {"data": {data_key: f"{base_url}/{bucket}/{prefix}/{img_name}.png"}}



from enum import Enum
from typing import Callable
import torch.nn as nn

from mmm.neural.TorchModule import TorchModule


class ActivationFn(str, Enum):
    Identity = "identity"
    Sigmoid = "sigmoid"
    LogSigmoid = "logsigmoid"
    Tanh = "tanh"
    ReLU = "relu"
    LeakyReLU = "leakyrelu"
    GeLU = "gelu"


fn_converter = {
    ActivationFn.Identity: nn.Identity,
    ActivationFn.LogSigmoid: nn.LogSigmoid,
    ActivationFn.Sigmoid: nn.Sigmoid,
    ActivationFn.Tanh: nn.Tanh,
    ActivationFn.ReLU: nn.ReLU,
    ActivationFn.LeakyReLU: nn.LeakyReLU,
    ActivationFn.GeLU: nn.GELU,
}


class ActivationFunctionConfig(TorchModule):
    fn_type: ActivationFn

    def build_instance(self, *args, **kwargs) -> nn.Module:
        return fn_converter[self.fn_type]()



from typing import cast
import numpy as np
import wandb
import logging
from sklearn import tree
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.svm import SVC, SVR
from sklearn.metrics import (
    accuracy_score,
    confusion_matrix,
    roc_auc_score,
    f1_score,
    mean_squared_error,
)
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LinearRegression
import torch
from scipy.stats import kendalltau


def evaluate_on_clf_arrays(
    X_train,
    y_train,
    X_val,
    y_val,
    class_names=None,
    confmattitle=None,
    confmatonlyval=True,
):
    sklearn_compatibles = {
        "randomforest": RandomForestClassifier,
    }
    softmax = torch.nn.Softmax(dim=1)
    scaler = StandardScaler()
    scaler.fit(X_train)
    X_train = scaler.transform(X_train)
    X_val = scaler.transform(X_val)
    imp = None
    for clf_name, clf_type in sklearn_compatibles.items():
        clf = clf_type()
        clf = clf.fit(X_train, y_train)
        y_pred_train: np.ndarray = clf.predict(X_train)
        y_pred_val: np.ndarray = clf.predict(X_val)
        if clf_name == "svm":
            decision = torch.from_numpy(cast(SVC, clf).decision_function(X_val))
            # logging.info(f"{decision.shape}")
            if len(decision.shape) == 1:
                decision = decision.unsqueeze(0)
            y_pred_proba: np.ndarray = softmax(decision).squeeze().numpy()
        else:
            y_pred_proba: np.ndarray = clf.predict_proba(X_val)
            if y_pred_proba.shape[-1] == 2:
                y_pred_proba = y_pred_proba[:, -1]
        res = {
            f"{clf_name}_train_acc": accuracy_score(y_true=y_train, y_pred=y_pred_train),
            # f"{clf_name}_train_confmatarr": confusion_matrix(y_true=y_train, y_pred=y_pred_train),
            f"{clf_name}_val_acc": accuracy_score(y_true=y_val, y_pred=y_pred_val),
            # f"{clf_name}_val_confmatarr": confusion_matrix(y_true=y_val, y_pred=y_pred_val),
            # f"{clf_name}_val_roc_auc": roc_auc_score(y_true=y_val, y_score=y_pred_proba, multi_class="ovr"),
            f"{clf_name}_val_f1": f1_score(y_true=y_val, y_pred=y_pred_val, average="micro"),
            # f"{clf_name}_val_f1_macro": f1_score(y_true=y_val, y_pred=y_pred_val, average="macro"),
            # f"{clf_name}_val_f1_weighted": f1_score(y_true=y_val, y_pred=y_pred_val, average="weighted"),
        }
        if clf_name == "randomforest":
            imp = clf.feature_importances_
        if confmattitle is not None:
            logging.info(f"{confmattitle}_{clf_name}_val")
            if not confmatonlyval:
                res[f"{clf_name}_train_confmat"] = wandb.plot.confusion_matrix(
                    preds=list(y_pred_train.tolist()),
                    y_true=list(y_train.tolist()),
                    class_names=class_names,
                    title=f"{confmattitle}_{clf_name}_train",
                )
            res[f"{clf_name}_val_confmat"] = wandb.plot.confusion_matrix(
                preds=list(y_pred_val.tolist()),
                y_true=list(y_val.tolist()),
                class_names=class_names,
                title=f"{confmattitle}_{clf_name}_val",
            )

        yield res, imp


def evaluate_on_reg_arrays(X_train, y_train, X_val, y_val, class_names=None):
    sklearn_compatibles = {
        "randomforest": RandomForestRegressor,
    }
    scaler = StandardScaler()
    X_train = scaler.fit_transform(X_train)
    X_val = scaler.transform(X_val)
    for clf_name, clf_type in sklearn_compatibles.items():
        clf = clf_type()
        clf = clf.fit(X_train, y_train)
        y_pred_train: np.ndarray = clf.predict(X_train)
        y_pred_val: np.ndarray = clf.predict(X_val)
        # y_proba_train: np.ndarray = clf.predict_proba(X_train)
        # y_proba_val: np.ndarray = clf.predict_proba(X_val)
        res = {
            # f"{clf_name}_train_kendalltau": kendalltau(x=y_train, y=y_pred_train),
            f"{clf_name}_train_mse": mean_squared_error(y_true=y_train, y_pred=y_pred_train),
            f"{clf_name}_val_mse": mean_squared_error(y_true=y_val, y_pred=y_pred_val),
            # f"{clf_name}_val_auc": roc_auc_score(y_true=y_val, y_score=y_pred_val)
        }
        imp = None
        if clf_name == "randomforest":
            imp = clf.feature_importances_

        yield res, imp



"""
Convenience utilities for testing without plausible training.

For a multi-task simulation dataset which can be trained on, take a look at the shape datasets.
"""

from typing import Any, Dict, Tuple, Optional
import torch
from torch.utils.data import Dataset

from mmm.data_loading.ClassificationDataset import ClassificationDataset
from mmm.data_loading.TrainValCohort import TrainValCohort
from mmm.mtl_modules.tasks.ClassificationTask import ClassificationTask


class ClassificationMockupDataset(Dataset):
    """
    Dataset of length N.
    """

    def __init__(self, N: int, image_shape=(1, 64, 64), return_real_task=False) -> None:
        super().__init__()
        self.N, self.image_shape = N, image_shape
        self.images = torch.ones(N, *image_shape)
        self.classes = torch.ones(N, dtype=torch.long)
        self.return_real_task = return_real_task
        self.class_names = ["No", "Yes"]
        for i in range(N, 2):
            self.images[i, 0] = torch.ones(image_shape[1], image_shape[2]) * 5
            self.classes[i] = 5

    def __len__(self):
        return self.N

    def __getitem__(self, index) -> Dict[str, Any]:
        if self.return_real_task:
            return {
                "image": self.images[index],
                "class": self.classes[index].unsqueeze(0),
            }
        else:
            class_code = int(index > (self.N // 2))
            return {
                "image": torch.ones(self.image_shape) * index,
                "class": torch.Tensor([class_code]),
            }

    @staticmethod
    def build_cohort(train_n: int, val_n: Optional[int]) -> TrainValCohort:
        train_src = ClassificationMockupDataset(train_n, image_shape=(1, 4, 4))
        val_src = ClassificationMockupDataset(val_n) if val_n is not None else None
        return TrainValCohort(
            TrainValCohort.Config(batch_size=(2, 2), shuffle_loaders=(False, False), num_workers=0),
            ClassificationDataset(train_src, class_names=train_src.class_names),
            (ClassificationDataset(val_src, class_names=val_src.class_names) if val_src is not None else None),
        )

    @classmethod
    def build_classification_task(cls, task_name: str, train_n: int, val_n: Optional[int]) -> ClassificationTask:
        cohort = cls.build_cohort(train_n, val_n)

        return ClassificationTask(
            64,
            # ["a", "b"],
            ClassificationTask.Config(
                module_name=task_name,
            ),
            cohort,
        )



from __future__ import annotations
from typing import Any, List, Optional, Callable, Dict, Tuple, TypeVar, Iterator

import torch
from torch.utils.data import Dataset

from .MTLDataset import MTLDataset, SrcCaseType, InvalidCaseError


class MultipleInstanceDataset(MTLDataset):
    """
    - bag: list of dicts, where each dict must contain the keys specified by instance_case_keys
    - label: single label for the bag given by the integer index of the class name in class_names
    """

    def __init__(
        self,
        src_ds: Dataset[SrcCaseType],
        provide_latent_rep: bool,
        class_names: List[str],
        src_transform: Optional[Callable[[SrcCaseType], Dict[str, Any]]] = None,
        batch_transform: Optional[Callable[[SrcCaseType], Dict[str, Any]]] = None,
        verify_cases: int = 0,
    ) -> None:
        self.class_names = class_names
        self.provide_latent_rep = provide_latent_rep
        self.src_transform = src_transform
        self.batch_transform = batch_transform
        self.instance_keys = ["image", "position"]
        super().__init__(src_ds, ["bag", "label", "meta"], [], None, None, None, verify_cases)

    def verify_case_by_index(self, index: int) -> Dict[str, Any]:
        case = super().verify_case_by_index(index)
        assert isinstance(case["bag"], list)
        if not self.provide_latent_rep:
            for instance in case["bag"]:
                missing_keys = [k for k in self.instance_keys if k not in instance]
                if missing_keys:
                    raise InvalidCaseError(
                        self,
                        -1,
                        [f"{k} missing in case {instance}" for k in missing_keys],
                    )
        else:
            assert isinstance(case["bag"], torch.Tensor), "Bag must be a tensor if provided as latent representation."

        return case

    def get_untransformed_case(self, index: int) -> Any:
        case = self.src_ds[self._indices[index]]
        if not self.provide_latent_rep:
            if self.src_transform is not None:
                for idx, img in enumerate(case["bag"]):
                    case["bag"][idx] = self.src_transform(img)
        return self.data_stripper(case)

    def __getitem__(self, index: int) -> Dict[str, Any]:
        src_case = self.get_untransformed_case(index) if self.cache_folder is None else self.get_case_from_cache(index)

        # The cohort may use these to apply random augmentations
        if not self.provide_latent_rep:
            if self.batch_transform is not None:
                for idx, img in enumerate(src_case["bag"]):
                    src_case["bag"][idx] = self.batch_transform(img)

        return src_case

    def __iter__(self) -> Iterator[Dict[str, Any]]:
        for d in iter(self.src_ds):
            if not self.provide_latent_rep:
                if self.src_transform is not None:
                    for idx, img in enumerate(d["bag"]):
                        d["bag"][idx] = self.src_transform(img)

                if self.batch_transform is not None:
                    for idx, img in enumerate(d["bag"]):
                        d["bag"][idx] = self.batch_transform(img)

            d = self.data_stripper(d)  # type: ignore
            yield d



import logging
import math
import random
from typing import (
    Literal,
    Optional,
    Callable,
    List,
    Tuple,
    Dict,
    Any,
)
import os
from zipfile import ZipFile
from shutil import unpack_archive
import requests
from pathlib import Path

from tqdm import tqdm
from typing import List, Optional, Tuple, Dict, Any
from pathlib import Path

import torch
from torch.utils.data import get_worker_info, Dataset
import collections
import re
import numpy as np

from .DetectionDataset import DetectionDataset
from .SemSegDataset import SemSegDataset


np_str_obj_array_pattern = re.compile(r"[SaUO]")


def batch_concat_with_meta(batch: List[Dict[str, Any]]) -> Dict[str, Any]:
    # Build the list of metainfo
    meta_dicts: List[Dict] = [x for item in batch for x in item["meta"]]
    for case_dict in batch:
        if "meta" in case_dict:
            case_dict.pop("meta")

    # Build the training-relevant batch
    res = batch_concat(batch)

    # Reappend the meta info as list
    res["meta"] = meta_dicts  # type: ignore
    return res  # type: ignore


def batch_concat(batch):
    """
    Adapted from https://github.com/pytorch/pytorch/blob/master/torch/utils/data/_utils/collate.py.

    For datasets which return batches of variable size the function can be used to overwrite collate_fn
    in a dataloader.
    """
    elem = batch[0]
    elem_type = type(elem)

    if isinstance(elem, torch.Tensor):
        if elem.numel() == 0:
            tensornums = [bool(t.numel()) for t in batch]
            # assert True in tensornums, "All tensors empty for batch concat"
            if not True in tensornums:
                logging.info("All Tensors empty for batch concat :(")
            else:
                elem = batch[tensornums.index(True)]
                elem_type = type(elem)

        out = None
        if get_worker_info() is not None:
            # If we're in a background process, concatenate directly into a
            # shared memory tensor to avoid an extra copy
            numel = sum(x.numel() for x in batch)
            storage = elem.storage()._new_shared(numel)
            out = elem.new(storage).resize_(sum([t.size(0) for t in batch]), *list(elem.size()[1:]))
        return torch.concat(batch, out=out)
    elif elem_type.__module__ == "numpy" and elem_type.__name__ != "str_" and elem_type.__name__ != "string_":
        if elem_type.__name__ == "ndarray" or elem_type.__name__ == "memmap":
            # array of string classes and object
            if np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                raise Exception(
                    f"default_collate: batch must contain tensors, "
                    + "numpy arrays, numbers, dicts or lists; found: {elem.dtype}"
                )

            return batch_concat([torch.as_tensor(b) for b in batch])
        elif elem.shape == ():  # scalars
            return torch.as_tensor(batch)
    elif isinstance(elem, float):
        return torch.tensor(batch, dtype=torch.float64)
    elif isinstance(elem, int):
        return torch.tensor(batch)
    elif isinstance(elem, collections.abc.Mapping):
        try:
            return elem_type({key: batch_concat([d[key] for d in batch]) for key in elem})
        except TypeError:
            # The mapping type may not support `__init__(iterable)`.
            return {key: batch_concat([d[key] for d in batch]) for key in elem}
    elif isinstance(elem, tuple) and hasattr(elem, "_fields"):  # namedtuple
        return elem_type(*(batch_concat(samples) for samples in zip(*batch)))
    elif isinstance(elem, collections.abc.Sequence):
        # check to make sure that the elements in batch have consistent size
        it = iter(batch)
        elem_size = len(next(it))
        if not all(len(elem) == elem_size for elem in it):
            raise RuntimeError("each element in list of batch should be of equal size.")
        transposed = list(zip(*batch))  # It may be accessed twice, so we use a list.

        if isinstance(elem, tuple):
            return [batch_concat(samples) for samples in transposed]  # Backwards compatibility.
        else:
            try:
                return elem_type([batch_concat(samples) for samples in transposed])
            except TypeError:
                # The sequence type may not support `__init__(iterable)` (e.g., `range`).
                return [batch_concat(samples) for samples in transposed]

    raise TypeError(batch_concat.format(elem_type))


def download_and_extract_archive(
    target: Path,
    ds_download_link: str,
    archive_type: Literal["infer", "zip", "targz"] = "infer",
    delete_archive_after=True,
    streaming_chunk_size=8192,
) -> Path:
    """
    Given a parent folder such as root / "your_dataset" and a download link, downloads and extracts archived data.

    Returns the path to the root directory in the archive.
    """
    logging.warn(f"Deprecated, use `from torchvision.datasets.utils import download_and_extract_archive`")
    if not target.exists():
        target.mkdir()

    if archive_type == "infer":
        if ds_download_link.lower().endswith(".zip"):
            archive_type = "zip"
        elif ds_download_link.lower().endswith(".tar.gz"):
            archive_type = "targz"
        else:
            raise Exception(f"Cannot infer archive type from {ds_download_link}")

    assert target.is_dir(), f"target {target} should be a directory"
    archive_path = target / "ds_archive"
    if not archive_path.exists():
        print(f"Downloading {ds_download_link} into {archive_path}")
        with open(archive_path, "wb") as f:
            resp = requests.get(ds_download_link, stream=True)
            for chunk in tqdm(resp.iter_content(chunk_size=streaming_chunk_size)):
                f.write(chunk)

    if archive_type == "zip":
        with ZipFile(archive_path, "r") as zip_ref:
            dirs = [target / d.filename for d in zip_ref.infolist() if d.is_dir()]
            if not dirs[0].exists():
                print(f"Unzipping {archive_path} into {target}")
                zip_ref.extractall(target)
    elif archive_type == "targz":
        if len(os.listdir(target)) == 1:
            unpack_archive(archive_path, extract_dir=target, format="gztar")
    else:
        raise Exception(f"Cannot work with archive type {archive_type}")

    if delete_archive_after:
        print(f"Removing the archive {archive_path}")
        os.remove(archive_path)

    print(f"Preparing data done")
    return target


class TransformedSubset(Dataset):
    """
    Transforms an existing dataset and filters using the given indices.

    Only applies indices or transform if not None.

    Known problem:
    This will use the methods of the supertype instead of your actual child type
    """

    def __init__(
        self,
        source_ds: Dataset,
        indices: Optional[List[int]] = None,
        transform: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = None,
    ):
        self.source_ds: Dataset = source_ds
        self.transform = transform
        if not indices:
            indices = list(range(len(self.source_ds)))  # type: ignore
        self.indices = indices

    # Only invoked if the attribute was not found on the actual object!
    def __getattr__(self, attr):
        return getattr(self.source_ds, attr)

    def __len__(self):
        return len(self.indices)

    def __getitem__(self, index):
        x = self.source_ds.__getitem__(self.indices[index])
        if self.transform is not None:
            x = self.transform(x)
        return x


def train_val_split_class_dependent(
    indices: List[int],
    classes: List[int],
    perc: float = 0.7,
    amount: int = 0,
    allow_imbalance: bool = False,
    deterministic_seed: int = -1,
) -> Tuple[List[int], List[int]]:
    """
    Given a list of indices and class indices, split the list into train indices and val indices.
    For train indices, make sure that all classes are represented at least once.

    Not a good use if there are 2 or 3 examples of one class and 100 of a different class.
    """
    class_names, count = np.unique(classes, return_counts=True)

    if amount > 0:
        train_len = amount
        val_len = 0
        desired_train_len = amount * len(class_names)
    else:
        min_fraction = min(count) / len(indices)
        desired_train_len = int(perc * len(indices))

        train_len = int((perc * min_fraction) * len(indices))
        val_len = int(((1 - perc) * min_fraction) * len(indices))

    train = []
    val = []
    used_idxs = []

    if deterministic_seed > 0:
        state = np.random.get_state()
        np.random.seed(deterministic_seed)

    for i in class_names:
        class_idx = [idx for idx in indices if classes[idx] == i]
        train_class_tmp = np.random.choice(class_idx, train_len, replace=False)
        for idx in train_class_tmp:
            train.append(idx)
            used_idxs.append(idx)
            class_idx.remove(idx)

        val_class_tmp = np.random.choice(class_idx, val_len, replace=False)

        for idx in val_class_tmp:
            used_idxs.append(idx)
            val.append(idx)

    unused_idx = [x for x in indices if x not in used_idxs]
    if allow_imbalance:
        while len(train) < desired_train_len:
            random_idx = np.random.choice(unused_idx, replace=False)
            train.append(random_idx)
            used_idxs.append(random_idx)
            unused_idx = [x for x in indices if x not in used_idxs]

    for rest in unused_idx:
        val.append(rest)

    if deterministic_seed > 0:
        np.random.set_state(state)

    return train, val


def train_val_split(
    indices: List[int],
    perc: float = 0.7,
    seed: Optional[int] = None,
    ensure_representativenumber_for_classes: Optional[List[int]] = None,
    min_representatives_per_class=1,
) -> Tuple[List[int], List[int]]:
    """
    Given a list of indices, splits the indices into (train_indices, val_indices).

    min_representatives_per_class is not respected perfectly.

    >>> from mmm.data_loading.utils import train_val_split
    >>> [len(split_indices) for split_indices in train_val_split(range(10), 0.51)]
    [6, 4]
    >>> train_classes = [0, 1, 2, 3, 3, 3, 3, 3]
    >>> a, b = train_val_split(list(range(len(train_classes))), perc=0.4, seed=0, ensure_representativenumber_for_classes=train_classes)
    >>> a
    [3, 0, 1, 2]
    >>> b
    [4, 5, 6, 7]
    """
    temp_random = random.Random(seed) if seed is not None else random
    unique = np.unique(np.asarray(ensure_representativenumber_for_classes))
    train_indices = temp_random.sample(indices, max(math.ceil(len(indices) * perc), len(unique)))

    if ensure_representativenumber_for_classes is not None:
        classes_pushed = []
        train_classes = [ensure_representativenumber_for_classes[i] for i in train_indices]

        for class_id in set(ensure_representativenumber_for_classes):
            # Make sure class_id is represented by train_indices
            while train_classes.count(class_id) < min_representatives_per_class:
                train_indices.append(ensure_representativenumber_for_classes.index(class_id))
                train_classes.append(class_id)
                classes_pushed.append(class_id)

                # Remove one of the overrepresented class
                overrepresented_class_id = max(set(train_classes), key=train_classes.count)
                assert (
                    overrepresented_class_id not in classes_pushed
                ), f"Impossible! Increase train perc {classes_pushed}"
                index_of_overrepresented_class_id = train_classes.index(overrepresented_class_id)
                train_indices.pop(index_of_overrepresented_class_id)
                train_classes.pop(index_of_overrepresented_class_id)

        if classes_pushed:
            logging.debug(f"Pushed validation classes into train after random sampling: {classes_pushed}")

    val_indices = set(indices) - set(train_indices)
    if not val_indices:
        val_indices = [train_indices.pop(0)]
        logging.warn(f"Not enough cases for train_val_split, reshuffled to {val_indices=}, {train_indices=}")
    return list(train_indices), list(val_indices)


def semseg_from_detect_ds(ds: DetectionDataset):
    """
    Converts detection bboxed to a segmentation map of size HxW
    """

    def convert_detectcase_to_semseg(detectcase: Dict[str, Any]) -> Dict[str, Any]:
        assert "image" in detectcase and "boxes" in detectcase and "labels" in detectcase

        mask = torch.zeros(detectcase["image"].shape)

        for idx, box in enumerate(detectcase["boxes"].tolist()):
            mask[box[0] : box[2], box[1] : box[3]] = detectcase["label"][idx]

        return {"image": detectcase["image"], "label": mask.long()}

    return SemSegDataset(
        src_ds=ds,
        src_transforms=convert_detectcase_to_semseg,
        class_names=ds.vis_classes,
    )



from typing import Dict, Any
import torch
import torchvision
import cv2 as cv
from monai.transforms import Transform


class RandomLineAugmentation(Transform):
    def __call__(self, data: torch.Tensor):
        """
        Expect a normlized RGB Torch Tensor
        """

        _, data_x, data_y = data.squeeze().shape
        # generate random points (without replacement)
        rnd_x = torch.randperm(data_x)[:2]
        rnd_y = torch.randperm(data_y)[:2]

        # transfrom into cv2 readable landmarks
        pt1 = (rnd_x[0].item(), rnd_y[0].item())
        pt2 = (rnd_x[1].item(), rnd_y[1].item())

        # create artifact mask
        artifacts = torch.ones((data_x, data_y, 1), dtype=torch.uint8).numpy() * 255.0
        line_mask = cv.line(artifacts, pt1, pt2, color=(0, 0, 0), thickness=data_x // 32) / 255.0

        rnd_blurr_divider = torch.randint(15, 20, size=(1,)).item()
        # (data_x//rnd_blurr_divider, data_y//rnd_blurr_divider))
        blurred_mask = cv.blur(line_mask, (data_x // rnd_blurr_divider, data_y // rnd_blurr_divider))

        # multiply input img with mask to create artifact
        return data * blurred_mask


class RandomBlobAugmentation(Transform):
    def __call__(self, data: torch.Tensor):
        _, data_x, data_y = data.squeeze().shape
        # generate random points
        min_radius = data_x // 30
        rnd_x = torch.randint(min_radius, data_x - min_radius, size=(1,))
        rnd_y = torch.randint(min_radius, data_y - min_radius, size=(1,))
        center = (rnd_x.item(), rnd_y.item())
        rnd_size_divider = torch.randint(7, 20, size=(1,)).item()

        artifacts = torch.ones((data_x, data_y, 1), dtype=torch.uint8).numpy() * 255.0
        circle_mask = (
            cv.circle(
                artifacts,
                center,
                radius=data_x // rnd_size_divider,
                color=(0, 0, 0),
                thickness=-1,
            )
            / 255.0
        )

        rnd_blurr_divider = torch.randint(15, 20, size=(1,)).item()
        blurred_mask = cv.blur(circle_mask, (data_x // rnd_blurr_divider, data_y // rnd_blurr_divider))
        return data * blurred_mask


class RandomTearAugmentation(Transform):
    def __call__(self, data: torch.Tensor):
        _, data_x, data_y = data.squeeze().shape
        data = data.squeeze()
        # generate random points
        x_shift = data_x // 30
        rnd_x, _ = torch.sort(torch.randperm(data_x - x_shift)[:2])
        rnd_y, _ = torch.sort(torch.randperm(data_y)[:2])
        while rnd_x[1] - rnd_x[0] < 10:
            rnd_x, _ = torch.sort(torch.randperm(data_x - x_shift)[:2])

        crop = data[:, rnd_x[0] : rnd_x[1], rnd_y[0] : rnd_y[1]]
        mask = torch.zeros((1, data_x, data_y))
        mask[:, rnd_x[0] : rnd_x[1], rnd_y[0] : rnd_y[1]] = 1
        tile_replacement = torch.ones(crop.shape)
        crop = torchvision.transforms.Resize(((crop.shape[1] // 3), crop.shape[2]))(crop)

        tile_replacement[:, : crop.shape[1], :] = crop[:, :, :]

        data[:, rnd_x[0] : rnd_x[1], rnd_y[0] : rnd_y[1]] = tile_replacement[:, :, :]
        return data.float()


class RandomArtifactAugmentation(Transform):
    def __init__(self, probability=0.7) -> None:
        self.artifact_augs = [
            # RandomTearAugmentation(),
            RandomLineAugmentation(),
            RandomBlobAugmentation(),
        ]
        self.p = probability

    def __call__(self, data: Dict[str, Any]):
        rnd = torch.rand(len(self.artifact_augs))
        for i in range(len(self.artifact_augs)):
            if rnd[i] < self.p:
                data["image"] = self.artifact_augs[i](data["image"])
        data["image"] = data["image"].float()
        return data



from typing import Literal, Any
import re
import logging
import random
from pydantic import Field
from copy import deepcopy
from shapely.geometry import Polygon
import numpy as np
from tiffslide import TiffSlide
import torch
import torchvision.transforms.functional as F

from mmm.BaseModel import BaseModel
from mmm.bucketizing import BucketConfig, Bucket

from .GeoAnno import GeoAnno
from .utils import (
    move_anno_to_window,
    rasterize_annotations,
    rasterize_multilabel_annotations,
    extract_detection_labels,
)

ObjectLabelType = Literal["det", "seg", "mlsemseg"]


class GeojsonObjWindows(BaseModel):
    """
    Each WSI is treated as a supercase with windows as subcases.
    This class offers a parameterized function for extracting windows per WSI.

    Enables segmentation, detection and multilabel classification.
    """

    minimum_pixels_level: int = Field(
        default=15 * 15,
        description="An annotation must represent at least this many pixels in the level to be included.",
    )
    filter_larger_than_fraction: float = Field(
        default=0.95,
        description="If an annotation is larger than this fraction of the window, it will be excluded.",
    )
    output_patch_size: tuple[int, int] = 1024, 1024
    minimum_visible_fraction: float = 0.2
    buckets: BucketConfig = BucketConfig(buckets=[Bucket(expression=f".*", description="all")])
    best_area_per_patch: float = 0.5
    target_label: list[ObjectLabelType] = ["det"]
    prefill_semsegmask_value: int = -1

    stuff_matcher: str | None = Field(
        default=None,
        description="Stuff is rasterized with in the back and not filtered by visibility criteria.",
    )

    def _pick_good_level(self, for_anno: GeoAnno, downsample_factors: list[float]) -> int | None:
        wind_wh = {
            level: (
                self.output_patch_size[0] * factor,
                self.output_patch_size[1] * factor,
            )
            for level, factor in enumerate(downsample_factors)
        }
        _, l0_anno_size = for_anno.get_enclosing_region()
        areas_per_level = {
            level: (
                for_anno.shape.area / (downsample_factor**2),
                l0_anno_size[0] / downsample_factor,
                l0_anno_size[1] / downsample_factor,
            )
            for level, downsample_factor in enumerate(downsample_factors)
        }
        # Collect the probability for each level to select it as the best level
        good_levels = {
            level: 1 - np.abs(self.best_area_per_patch - ((w * h) / (wind_wh[level][0] * wind_wh[level][1])))
            for level, (ar, w, h) in areas_per_level.items()
            if (
                (ar > self.minimum_pixels_level)
                and (ar / (wind_wh[level][0] * wind_wh[level][1]) < self.filter_larger_than_fraction)
                and (w < self.output_patch_size[0])
                and (h < self.output_patch_size[1])
            )
        }
        if not good_levels:
            # If there is no really good level for this annotation, skip the window
            logging.debug(f"Found no good window for {for_anno}")
            return None
        # Pick one of the levels that are suitable for this annotation:
        window_level = random.choices(list(good_levels.keys()), weights=list(good_levels.values()), k=1)[0]
        return window_level

    def select_good_window(
        self, for_anno: GeoAnno, downsample_fac: float, wsi_rows: int, wsi_cols: int
    ) -> tuple[Polygon, int, int, Polygon] | None:
        l0_anno_location, l0_anno_size = for_anno.get_enclosing_region()
        l0_window_rows, l0_window_cols = (
            self.output_patch_size[0] * downsample_fac,
            self.output_patch_size[1] * downsample_fac,
        )
        max_x = min(wsi_rows - l0_window_rows, l0_anno_location[0])
        min_x = max(0, l0_anno_location[0] - (l0_window_rows - l0_anno_size[0]))
        max_y = min(wsi_cols - l0_window_cols, l0_anno_location[1])
        min_y = max(0, l0_anno_location[1] - (l0_window_cols - l0_anno_size[1]))
        if min_x > max_x or min_y > max_y:
            logging.debug(f"The annotation {for_anno} is too large for the selected level")
            return None
        l0_x, l0_y = random.randint(int(min_x), int(max_x)), random.randint(int(min_y), int(max_y))
        l0_window = Polygon(
            [
                (l0_x, l0_y),
                (l0_x + l0_window_rows, l0_y),
                (l0_x + l0_window_rows, l0_y + l0_window_cols),
                (l0_x, l0_y + l0_window_cols),
            ]
        )

        l_x, l_y = int(l0_x / downsample_fac), int(l0_y / downsample_fac)
        l_window = Polygon(
            [
                (l_x, l_y),
                (l_x + self.output_patch_size[0], l_y),
                (l_x + self.output_patch_size[0], l_y + self.output_patch_size[1]),
                (l_x, l_y + self.output_patch_size[1]),
            ]
        )
        return l0_window, l0_x, l0_y, l_window

    def transform_wsi_to_windows(self, supercase: dict[str, Any]) -> list[dict[str, Any]]:
        """
        Supercases need to have at least the keys:

        - wsi: TiffSlide
        - annos: List[GeoAnno]

        and should have a meta key containing a custom dictionary.

        This method yields windows with the annotations in them.
        """
        res = []
        for window_dict in self._iterate_windows(supercase):
            slide, l0_x, l0_y, window_level, annos_in_window, buckets, anno_labels = (
                window_dict["slide"],
                window_dict["l0_x"],
                window_dict["l0_y"],
                window_dict["window_level"],
                window_dict["annos_in_window"],
                window_dict["buckets"],
                window_dict["anno_labels"],
            )
            try:
                img_region = slide.read_region(
                    location=(l0_x, l0_y),
                    level=window_level,
                    size=self.output_patch_size,
                )
            except Exception as e:
                logging.error(
                    f"Region read error {e=}\n"
                    + "for {l0_x}, {l0_y}, {window_level}, {self.output_patch_size} in {window_dict['slide']=}"
                )
                continue
            subcase_dict: dict[str, Any] = {
                "image": F.to_tensor(img_region),
                "meta": {
                    "level": window_level,
                    "window_location": (l0_x, l0_y),
                    "classnames": anno_labels,
                },
            }

            for label_type in self.target_label:
                self._extract_labels(subcase_dict, label_type, annos_in_window, buckets)

            if "meta" in supercase:
                subcase_dict["meta"]["supercase"] = supercase["meta"]
            res.append(subcase_dict)

        return res

    def _is_window_perfect_for_anno(self, l0_window: Polygon, l0_anno: GeoAnno, downsample_fac: float) -> bool:
        # Not too big or only partly visible:
        by_overlap = l0_anno.overlaps_with(l0_window, 0.99)
        # Not too small:
        by_pixel_num = (l0_anno.shape.area / (downsample_fac**2)) > self.minimum_pixels_level
        return by_overlap and by_pixel_num

    def is_anno_useful_for_window(self, l0_window: Polygon, l0_anno: GeoAnno, downsample_fac: float) -> bool:
        return l0_anno.overlaps_with(l0_window, self.minimum_visible_fraction)

    def _is_stuff(self, anno: GeoAnno) -> bool:
        # Match using regex
        if self.stuff_matcher is None:
            return False

        return re.match(self.stuff_matcher, anno.get_class_name()) is not None

    def _iterate_windows(self, supercase: dict[str, Any]) -> list[dict[str, Any]]:
        slide: TiffSlide = TiffSlide(supercase["wsi"])

        wsi_rows, wsi_cols = slide.level_dimensions[0][0], slide.level_dimensions[0][1]
        # This buffer will be used to create windows
        all_annos: list[GeoAnno] = [anno for anno in supercase["annos"]]
        # Only use index in this list
        annos_left: list[int] = [i for i, _ in enumerate(all_annos)]
        random.shuffle(annos_left)

        if self.stuff_matcher is not None:
            # Put stuff in the back of the list to give objects the chance for a good window first
            stuff_annos = filter(lambda i: self._is_stuff(all_annos[i]), annos_left)
            thing_annos = filter(lambda i: not self._is_stuff(all_annos[i]), annos_left)
            annos_left = list(thing_annos) + list(stuff_annos)

        while annos_left:
            # The perfect window for this annotation is selected
            main_annotation_index = annos_left.pop(0)
            main_annotation: GeoAnno = all_annos[main_annotation_index]

            # Find the level where the annotation makes up a maximum of 50% of the area
            window_level = self._pick_good_level(main_annotation, slide.level_downsamples)
            if window_level is None:
                continue
            downsample_fac = slide.level_downsamples[window_level]

            l0_window, l0_x, l0_y, l_window = self.select_good_window(
                main_annotation, downsample_fac, wsi_rows, wsi_cols
            )
            if l0_window is None:
                continue

            other_annos = []
            for all_anno_index, l0_anno in enumerate(all_annos):
                if all_anno_index != main_annotation_index:
                    # Determine if anno is well represented, partially represented or not represented in this window
                    if self._is_window_perfect_for_anno(l0_window, l0_anno, downsample_fac):
                        other_annos.append(deepcopy(l0_anno))
                        if all_anno_index in annos_left:
                            annos_left.remove(all_anno_index)
                    elif self.is_anno_useful_for_window(l0_window, l0_anno, downsample_fac):
                        # This annotation is not yet well represented, it should get another chance in its own window
                        # In consequence, it is not removed from the list of annotations left
                        other_annos.append(deepcopy(l0_anno))
                    elif self._is_stuff(l0_anno) and l0_anno.shape.intersection(l0_window).area > 0:
                        other_annos.append(deepcopy(l0_anno))

            other_annos.append(deepcopy(main_annotation))

            annos_in_window: list[GeoAnno] = list(
                map(
                    lambda anno: move_anno_to_window(anno, l0_window, l0_x, l0_y, downsample_fac),
                    other_annos,
                )
            )
            # Filter those that are too small or too big in this window
            annos_in_window = [
                anno
                for anno in annos_in_window
                if self._is_stuff(anno)
                or (
                    (anno.shape.area > self.minimum_pixels_level)
                    and (anno.shape.area / l_window.area < self.filter_larger_than_fraction)
                )
            ]
            if not annos_in_window:
                # This should never happen because the window is extracted to fit at least one annotation
                logging.warning("Somehow I got a window without useful annotations")

            # Sort the stuff annotations into the beginning of the list to ensure they do not overwrite things
            annos_in_window.sort(key=lambda anno: not self._is_stuff(anno))

            anno_labels = [anno.get_class_name() for anno in annos_in_window]
            buckets: list[tuple[str, int]] = [self.buckets.get_bucket_name(l.lower()) for l in anno_labels]
            yield dict(
                annos_in_window=annos_in_window,
                buckets=buckets,
                l0_window=l0_window,
                l_window=l_window,
                slide=slide,
                l0_x=l0_x,
                l0_y=l0_y,
                window_level=window_level,
                anno_labels=anno_labels,
            )

    def _extract_labels(
        self,
        subcase_dict: dict[str, Any],
        t: ObjectLabelType,
        annos_in_window: list[GeoAnno],
        buckets: list[tuple[str, int]],
    ) -> None:
        if t == "det":
            subcase_dict = extract_detection_labels(subcase_dict, annos_in_window, buckets, self.output_patch_size)
        elif t == "seg":
            subcase_dict["label"] = torch.from_numpy(
                rasterize_annotations(
                    self.output_patch_size[0],
                    self.output_patch_size[1],
                    annos_in_window,
                    torch.Tensor([x[1] for x in buckets]).long(),
                    unlabeled_value=self.prefill_semsegmask_value,
                )
            )
        elif t == "mlsemseg":
            subcase_dict["masks"] = torch.from_numpy(
                rasterize_multilabel_annotations(
                    self.output_patch_size[0],
                    self.output_patch_size[1],
                    annos_in_window,
                    torch.Tensor([x[1] for x in buckets]).long(),
                    len(self.buckets.buckets),
                    unlabeled_value=self.prefill_semsegmask_value,
                )
            )
        else:
            raise NotImplementedError



from __future__ import annotations
from enum import Enum


class DataSplit(Enum):
    """
    By the time of writing, the library is only opinionated about training.
    As a result, only train and validation split handling is directly integrated.
    If you want to integrate testing or other custom code directly into the training,
    you can use the trainer's hooks.

    In this enum the actual values matter because other components are allowed to use them for indexing.
    """

    train = 0
    val = 1

    @staticmethod
    def from_index(i: int) -> DataSplit:
        return DataSplit.val if i != 0 else DataSplit.train



import logging
import random
import json
from pydantic import Field
from typing import Dict, Tuple, Any, List, cast, Literal, Union, Optional

import wandb
import numpy as np
import torch
import torch.nn as nn

import torchvision.transforms.functional as F
from torchvision.transforms import Resize, InterpolationMode

from mmm.logging.type_ext import StepMetricDict

from mmm.mtl_modules.shared_blocks.PyramidDecoder import PyramidDecoder

from .MTLTask import MTLTask
from mmm.settings import mtl_settings
from mmm.utils import flatten_list_of_dicts
from mmm.mtl_modules.shared_blocks.SharedBlock import SharedBlock
from mmm.mtl_modules.shared_blocks.SharedModules import SharedModules
from mmm.data_loading.TrainValCohort import TrainValCohort
from mmm.data_loading.SemSegDataset import SemSegDataset
from mmm.neural.module_conversions import convert_2d_to_3d
from mmm.mtl_modules.shared_blocks.MTLDecoder import MTLDecoder
from mmm.neural.losses import SMPDiceFocalLoss2D

import segmentation_models_pytorch.metrics as smp_metrics


@torch.no_grad()
def get_stats_multilabel(
    output: torch.LongTensor,
    target: torch.LongTensor,
    ignore_value: int = -1,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Expects [B, C, *dims] tensors of predictions and targets.
    Not logits or scores!
    Returns tp,fp,fn,tn per sample per class.

    >>> # A batch with one image, with one class
    >>> from mmm.mtl_modules.tasks.SemSegTask import get_stats_multilabel; import torch
    >>> preds11 = torch.Tensor([[[[0, 1], [0, 1]]]]).long()
    >>> trues11 = torch.Tensor([[[[0, -1], [0, 1]]]]).long()
    >>> get_stats_multilabel(preds11, trues11, ignore_value=-1)
    (tensor([[1]]), tensor([[0]]), tensor([[0]]), tensor([[2]]))
    >>> preds = torch.Tensor([[[[0, 1], [0, 1]], [[1, 1], [0, 1]]], [[[0, 0], [1, 1]], [[0, 0], [0, 1]]]]).long()
    >>> trues = torch.Tensor([[[[-1, -1], [1, 1]], [[0, 1], [0, -1]]], [[[0, -1], [-1, 0]], [[0, -1], [0, -1]]]]).long()
    >>> # True negatives of the second image of the second class
    >>> get_stats_multilabel(preds, trues, ignore_value=-1)[3][1, 1]
    tensor(2)
    """
    output, target = output.clone().long(), target.clone().long()

    if ignore_value is not None:
        mask = target != ignore_value
        output, target = output * mask, target * mask

    tp = (output * target).sum(dim=(-1, -2))
    fp = (output * (1 - target)).sum(dim=(-1, -2))
    fn = ((1 - output) * target).sum(dim=(-1, -2))
    # Ignored values will appear in this count
    tn = ((1 - output) * (1 - target)).sum(dim=(-1, -2))

    # Subtract the number of ignored values from tn
    tn -= (mask == False).long().sum(dim=(-1, -2))

    return tp, fp, fn, tn


def get_single_metrics(
    pred_mask: torch.Tensor,
    true_mask: torch.Tensor,
    original_shape: Tuple[int, int],
    class_names: List[str],
    ignore_index: int = -1,
    mode: Literal["multiclass", "multilabel"] = "multiclass",
):
    """
    Expects single images, not batches!
    """
    # Sometimes the metrics are desired on the original image size, not the network output size
    if original_shape != pred_mask.shape[-2:]:
        if mode == "multiclass":
            assert len(pred_mask.shape) == len(original_shape)
            assert len(true_mask.shape) == len(original_shape)
        elif mode == "multilabel":
            assert len(pred_mask.shape) == len(original_shape) + 1
            assert len(true_mask.shape) == len(original_shape) + 1
        resize_function = Resize(size=original_shape, interpolation=InterpolationMode.NEAREST)
        _pred_mask = resize_function(torch.unsqueeze(pred_mask, dim=0))
        _true_mask = resize_function(torch.unsqueeze(true_mask, dim=0))
    else:
        _pred_mask: torch.LongTensor = torch.unsqueeze(pred_mask, dim=0)  # type: ignore
        _true_mask: torch.LongTensor = torch.unsqueeze(true_mask, dim=0)  # type: ignore

    # Sadly, smp can only compute metrics with ignore_index outside the valid class indices
    num_classes = len(class_names)
    if mode == "multiclass":
        # _pred_mask[_pred_mask == ignore_index] = -1
        # _true_mask[_true_mask == ignore_index] = -1
        # _pred_mask[_pred_mask > ignore_index] -= 1
        # _true_mask[_true_mask > ignore_index] -= 1
        # num_classes -= 1

        return smp_metrics.get_stats(_pred_mask, _true_mask, mode=mode, num_classes=num_classes, ignore_index=-1)
    elif mode == "multilabel":
        return get_stats_multilabel(_pred_mask, _true_mask, ignore_value=ignore_index)


def get_batch_metrics(
    pred_masks: torch.Tensor,
    true_masks: torch.Tensor,
    original_shapes: List[Tuple[int, int]],
    class_names: List[str],
    ignore_index: int = -1,
    mode: Literal["multiclass", "multilabel"] = "multiclass",
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Setting original shape is important because the metrics are better on a smaller image size!

    original shapes is a list of the shapes for the whole batch:
    """
    res = []
    # assert isinstance(pred_masks, torch.LongTensor) and isinstance(true_masks, torch.LongTensor)
    assert pred_masks.shape == true_masks.shape

    # if original_shapes is None:
    #     return smp_metrics.get_stats(
    #         pred_masks,
    #         true_masks,
    #         mode="multiclass", num_classes=len(class_names)
    #     )  # type: ignore (They annotated the tuple wrong)
    # self.args.batch_size cannot be used, because the last batch can be smaller than batch_size
    for index_in_batch in range(pred_masks.shape[0]):
        single_image_metrics = get_single_metrics(
            pred_masks[index_in_batch],
            true_masks[index_in_batch],
            tuple(original_shapes[index_in_batch]),
            class_names,
            ignore_index=ignore_index,
            mode=mode,
        )
        res.append(single_image_metrics)
    tp = torch.concat([x[0] for x in res])
    fp = torch.concat([x[1] for x in res])
    fn = torch.concat([x[2] for x in res])
    tn = torch.concat([x[3] for x in res])
    return tp, fp, fn, tn


class SemSegTask(MTLTask):
    """
    Semantic segmentation task. Uses a fixed loss function by adding focalloss and dice loss together.

    Values in the mask that are equal to -1 will be ignored in the loss and metrics.
    This value can also be predicted when the best predicted class is below the confidence threshold.

    For building the task-specific heads it needs the number of output channels of the decoder it should use.

    Expects a dataset with tuples (image, mask). For example:

    >>> from mmm.interactive import *
    >>> from mmm.data_loading.synthetic.shape_dataset import ShapeDataset, CanvasConfig
    >>> ds = ShapeDataset(ShapeDataset.Config())
    >>> ssds = data.SemSegDataset(
    ...     src_ds=ds,
    ...     src_transform=ShapeDataset.cohort_transform(),
    ...     class_names=["background"] + ds.get_class_names())
    >>> t = tasks.SemSegTask(
    ...     args=configs.SemSegTaskConfig(module_name="TestSegmentation"),
    ...     cohort=data.TrainValCohort(configs.TrainValCohortConfig(), ssds, val_ds=None),
    ...     for_decoder=blocks.PyramidDecoder(PyramidDecoder.Config(), [32, 64, 128], 32),
    ...     class_names=ssds.class_names
    ... )
    """

    class Config(MTLTask.Config):
        encoder_key: str = "encoder"
        decoder_key: str = "decoder"
        squeezer_key: str = "squeezer"
        decoder_type: Literal["pyramid", "mtldecoder"] = "pyramid"
        loss: Literal["dicefocal"] = "dicefocal"
        dropout: float = 0.0
        headdropout: float = 0.0
        head_kernel_size: int = Field(default=3, description="Controls the receptive field of the head.")
        mask_key: str = "label"
        crit: SMPDiceFocalLoss2D.Config = SMPDiceFocalLoss2D.Config()
        confidence_threshold: float = 0.5

    def __init__(
        self,
        class_names: List[str],
        for_decoder: Union[MTLDecoder, PyramidDecoder],
        args: Config,
        cohort: TrainValCohort[SemSegDataset],
    ):
        super().__init__(args, cohort)
        self.args: SemSegTask.Config
        self.class_names = class_names
        if cohort and self.class_names != cohort.datasets[0].class_names:
            logging.warning(
                f"Class names {self.class_names} do not match dataset class names {cohort.datasets[0].class_names}"
            )

        self.task_modules = self.build_head(for_decoder)
        self.crit = self.build_loss()

    def build_head(self, for_decoder: Optional[Union[MTLDecoder, PyramidDecoder]] = None):
        return nn.ModuleDict(
            {
                "segmentation_head": nn.Conv2d(
                    for_decoder.get_output_dim_per_pixel(),
                    len(self.class_names),
                    kernel_size=self.args.head_kernel_size,
                    padding=self.args.head_kernel_size // 2,
                ),
                "dropout": nn.ModuleList(
                    [(nn.Dropout2d(p=self.args.dropout) if self.args.dropout > 0 else nn.Identity()) for _ in range(4)]
                ),
                "headdropout": (nn.Dropout2d(p=self.args.headdropout) if self.args.headdropout > 0 else nn.Identity()),
            }
        )

    def build_loss(self):
        return SMPDiceFocalLoss2D(
            self.args.crit,
            mode="multiclass",
            ignore_index=mtl_settings.ignore_class_value,
        )

    def prepare_batch(self, batch: Dict[str, Any]) -> Any:
        batch["image"] = batch["image"].to(self.torch_device)
        spatial_dims = batch["image"].size()[2:]  # type: ignore
        assert False not in [
            dim % 32 == 0 for dim in spatial_dims
        ], f"All spatial dimensions need to be divisable by 32, but you have {spatial_dims=}"

        batch[self.args.mask_key] = batch[self.args.mask_key].to(self.torch_device)
        return batch

    def forward(self, x: Dict[str, Any], shared_blocks: Dict[str, SharedBlock]):
        feat = shared_blocks[self.args.encoder_key](x)

        if hasattr(self.args, "squeezer_key") and self.args.squeezer_key in list(shared_blocks.keys()):
            feat[-1], _ = shared_blocks[self.args.squeezer_key](feat)

        if self.args.dropout > 0.0:
            feat = [feat[0]] + [self.task_modules["dropout"][i](f) for i, f in enumerate(feat[1:])]  # type: ignore

        # The decoder computes from the feature pyramid and the image's hidden dimension an embedding per pixel
        if self.args.decoder_type == "mtldecoder":
            decoder: MTLDecoder = shared_blocks[self.args.decoder_key]  # type: ignore
            raise Exception("MTLDecoder requires conditioning")
            decoder_output = decoder.forward(feat, z)  # type: ignore
        elif self.args.decoder_type == "pyramid":
            decoder: PyramidDecoder = cast(PyramidDecoder, shared_blocks[self.args.decoder_key])
            decoder_output = decoder.forward(feat)
        else:
            raise Exception(f"Unknown decoder type: {self.args.decoder_type}")

        # task specific computation
        if self.args.headdropout > 0.0:
            decoder_output = self.task_modules["headdropout"](decoder_output)
        masks_logits = self.task_modules["segmentation_head"](decoder_output)
        return masks_logits

    def check_batch_for_skippability(self, batch: Dict[str, Any]) -> bool:
        mask = batch[self.args.mask_key]
        if mtl_settings.ignore_class_value is not None:
            # If there is an ignore_index, the batch can be skipped if everything should be ignored
            uniques = torch.unique(mask)
            if uniques.shape[0] == 1 and (uniques[0] == mtl_settings.ignore_class_value).item():
                logging.warn(
                    f"In {self.args.module_name}, a whole batch with size {mask.shape} had only ignore index {mtl_settings.ignore_class_value}."
                )
                return True
        return False

    def training_step(self, batch: Dict[str, Any], shared_blocks: SharedModules) -> Optional[Tuple[torch.Tensor, Dict]]:
        im = batch["image"]
        mask = batch[self.args.mask_key]
        # Skip the batch, if there is only the ignore index in the mask
        if self.check_batch_for_skippability(batch):
            return None

        metas = batch["meta"] if "meta" in batch else [{} for _ in range(im.shape[0])]

        y_hat = shared_blocks.forward(im, self.forward)

        if y_hat.shape != mask.shape:
            mask = F.resize(mask, y_hat.shape[-2:], interpolation=InterpolationMode.NEAREST)

        realtime_log = {}

        losses = self.crit(y_hat, mask)
        dice_loss = losses["dice"]
        focal_loss = losses["focal"]
        realtime_log.update({"dice": dice_loss.item(), "focal": focal_loss.item()})
        batch_loss = self.normalize_loss(dice_loss, focal_loss)

        # Keep on GPU because metrics can be computed on GPU
        im_detached, mask_detached, y_hat_detached = (
            im.detach(),
            mask.detach(),
            y_hat.detach(),
        )
        # y_hat_detached = torch.argmax(y_hat_detached.sigmoid(), dim=1)
        # y_hat_preds = self.probas_to_preds(
        #     out_probas=self.logits_to_probas(y_hat_detached),
        #     confidence_threshold=self.args.confidence_threshold if mtl_settings.ignore_class_value is not None else None,
        #     uncertainty_class=mtl_settings.ignore_class_value,
        # )
        y_hat_probas = self.logits_to_probas(y_hat_detached)
        y_hat_preds = self.probas_to_preds(out_probas=y_hat_probas)
        step_metrics = self._collect_stats_for_metrics(
            mask_detached,
            y_hat_preds,
            batch.get("original_shape", [im.shape[-2:] for _ in range(im.shape[0])]),
        )

        self.add_step_result(batch_loss.item(), step_metrics)

        realtime_log.update(self._visualize_preds(im_detached, mask_detached, y_hat_preds, step_metrics, metas))

        return batch_loss, realtime_log

    def normalize_loss(self, dice_loss: torch.Tensor, focal_loss: torch.Tensor) -> torch.Tensor:
        # For now, normalization like normal cross entropy seems to be sufficient
        return dice_loss + (focal_loss / np.log(len(self.class_names)))

    @torch.no_grad()
    def _collect_stats_for_metrics(self, mask_detached, y_hat_preds, for_shape):
        tp, fp, fn, tn = get_batch_metrics(
            y_hat_preds.long(),
            mask_detached.long(),
            for_shape,
            self.class_names,
            ignore_index=mtl_settings.ignore_class_value,
        )
        step_metrics: StepMetricDict = {
            "tp": tp.numpy(),
            "fp": fp.numpy(),
            "fn": fn.numpy(),
            "tn": tn.numpy(),
        }  # type: ignore
        return step_metrics

    @torch.no_grad()
    def _visualize_pred(self, img, mask, pred, im_iou, meta: Dict) -> wandb.Image:
        if pred.shape[-2:] != img.shape[-2:]:
            # Add channel dim, resize, remove channel dim:
            mask = F.resize(
                mask.unsqueeze(0),
                img.shape[-2:],
                interpolation=InterpolationMode.NEAREST,
            ).squeeze()
            pred = F.resize(
                pred.unsqueeze(0),
                img.shape[-2:],
                interpolation=InterpolationMode.NEAREST,
            ).squeeze()

        metastr = json.dumps(meta, default=str)
        classes = {i + 1: self.class_names[i] for i in range(len(self.class_names))}
        assert mtl_settings.ignore_class_value not in classes, "Unexpected ignore class value in class names!"
        classes.update({mtl_settings.ignore_class_value + 1: "ignored"})
        res = wandb.Image(
            img,
            masks={
                "predictions": {
                    "mask_data": pred.numpy() + 1,
                    "class_labels": classes,
                },
                "ground_truth": {
                    "mask_data": mask.numpy() + 1,
                    "class_labels": classes,
                },
            },
            caption=f"{self.class_names}\nIOU: {im_iou.numpy()}\n[{torch.min(img):.2f}, {torch.max(img):.2f}]\n{img.shape}\n{metastr}]",
        )
        return res

    @torch.no_grad()
    def _visualize_preds(self, train_ims, train_masks, preds, step_metrics: Dict, metas: List[Dict]) -> Dict:
        res = {}

        vis_n = min(self._takeout_vis_budget(), train_ims.size(0))
        if vis_n >= 1:
            batch_size = train_ims.size(0)
            wandb_ims = []
            for rand_index in random.sample(list(range(batch_size)), vis_n):
                im, mask, pred = (
                    train_ims[rand_index].cpu(),
                    train_masks[rand_index].cpu(),
                    preds[rand_index].cpu(),
                )
                im_iou = smp_metrics.iou_score(
                    torch.LongTensor(step_metrics["tp"][rand_index]),
                    torch.LongTensor(step_metrics["fp"][rand_index]),
                    torch.LongTensor(step_metrics["fn"][rand_index]),
                    torch.LongTensor(step_metrics["tn"][rand_index]),
                )
                wandb_ims.append(self._visualize_pred(im, mask, pred, im_iou, metas[rand_index]))

            res["preds"] = wandb_ims
        return res

    def log_epoch_metrics(self) -> Tuple[Dict[str, Any], str]:
        flat_metrics = flatten_list_of_dicts(self._step_metrics)

        task_log_dict = {}

        by_class_ious = smp_metrics.iou_score(
            torch.LongTensor(flat_metrics["tp"]),
            torch.LongTensor(flat_metrics["fp"]),
            torch.LongTensor(flat_metrics["fn"]),
            torch.LongTensor(flat_metrics["tn"]),
        ).mean(
            dim=0
        )  # iou_score returns iou of all images by all classes [IMAGES, CLASSES]
        if len(self.class_names) < mtl_settings.max_classes_detailed_logging:
            for i, class_name in enumerate(self.class_names):
                # if mtl_settings.ignore_class_value is not None and i == mtl_settings.ignore_class_value:
                #     continue
                # class_idx = (
                #     i - 1 if mtl_settings.ignore_class_value is not None and i > mtl_settings.ignore_class_value else i
                # )
                task_log_dict[f"{class_name}iou"] = by_class_ious[i]
        # if mtl_settings.ignore_class_value is not None:
        #     by_class_ious = np.delete(by_class_ious, mtl_settings.ignore_class_value, None)
        task_log_dict["meanclassiou"] = by_class_ious.mean()

        _, print_str = super().log_epoch_metrics()
        return (
            task_log_dict,
            f"{print_str} - mean-class-iou {task_log_dict['meanclassiou']}",
        )

    @torch.no_grad()
    def get_logits_for_inputs(
        self,
        input_batch: torch.Tensor,
        shared_blocks: Dict[str, SharedBlock],
    ) -> torch.Tensor:
        """
        Returns multiclass probabilities for each pixel in the input image using softmax.
        """
        input_batch = input_batch.to(shared_blocks[self.args.decoder_key].torch_device)
        logits = self.forward(input_batch, shared_blocks)  # type: ignore
        return logits

    @torch.no_grad()
    def logits_to_probas(self, logits: torch.Tensor, output_size: Optional[Tuple[int, int]] = None) -> torch.Tensor:
        """
        Converts logits to probabilities using softmax.

        If output_size is set, the probabilities will be resized to that size.

        It is not merged with get_logits_for_inputs because then the logits can be computed using ONNX.
        """
        probas = logits.softmax(dim=1)
        if output_size is not None:
            probas = F.resize(probas, list(output_size), interpolation=InterpolationMode.NEAREST_EXACT)
        return probas

    @torch.no_grad()
    def probas_to_preds(self, out_probas: torch.Tensor, pixel_threshold: float | None = None):
        """
        Computes predictions for softmax probabilities, e.g. computed using SemSegTask.static_inference.

        If confidence_threshold is set, all pixels that do not have a certain prediction above that threshold
        will be set to uncertainty_class.
        """
        if pixel_threshold is None:
            if self.args.confidence_threshold is None:
                pixel_threshold = None
            else:
                pixel_threshold = self.args.confidence_threshold

        max_values, prediction = torch.max(out_probas, dim=1)
        if pixel_threshold is not None:
            prediction[max_values < pixel_threshold] = mtl_settings.ignore_class_value
        return prediction


class MultiLabelSemSegTask(SemSegTask):
    """
    At the ignore index (value in this case), the training process will assume that the area is unlabeled.
    Both metrics and loss will ignore those pixels.
    """

    class Config(SemSegTask.Config):
        mask_key: str = "masks"

    def build_loss(self):
        return SMPDiceFocalLoss2D(
            self.args.crit,
            mode="multilabel",
            ignore_index=mtl_settings.ignore_class_value,
        )

    def check_batch_for_skippability(self, batch: Dict[str, Any]) -> bool:
        mask = batch[self.args.mask_key]
        if mtl_settings.ignore_class_value is not None:
            # If there is an ignore_index, the batch can be skipped if the whole ignore dimension is set to 1
            if torch.all(mask[...] == mtl_settings.ignore_class_value):
                logging.warn(
                    f"In {self.args.module_name}, a whole batch with size {mask.shape} had only ignore value {mtl_settings.ignore_class_value}."
                )
                return True
        return False

    @torch.no_grad()
    def _collect_stats_for_metrics(self, mask_detached, y_hat_detached, for_shape):
        tp, fp, fn, tn = get_batch_metrics(
            y_hat_detached.long(),
            mask_detached.long(),
            for_shape,
            self.class_names,
            ignore_index=mtl_settings.ignore_class_value,
            mode="multilabel",
        )
        step_metrics: StepMetricDict = {
            "tp": tp.cpu().numpy(),
            "fp": fp.cpu().numpy(),
            "fn": fn.cpu().numpy(),
            "tn": tn.cpu().numpy(),
        }  # type: ignore
        return step_metrics

    @torch.no_grad()
    def logits_to_probas(self, logits: torch.Tensor, output_size: Tuple[int, int] | None = None) -> torch.Tensor:
        probas = logits.sigmoid()
        if output_size is not None:
            probas = F.resize(probas, list(output_size), interpolation=InterpolationMode.NEAREST)
        return probas

    @torch.no_grad()
    def probas_to_preds(self, out_probas: torch.Tensor, overwrite_threshold: float | None = None):
        """
        Computes predictions for sigmoid probabilities, e.g. computed using SemSegTask.static_inference.

        If confidence_threshold is set, all pixels that do not have a certain prediction above that threshold
        will be set to uncertainty_class.
        Confidence threshold will also be used for the value where a class is considered true.
        """
        prediction = torch.zeros_like(out_probas)

        threshold = self.args.confidence_threshold if overwrite_threshold is None else overwrite_threshold

        prediction[out_probas > threshold] = 1
        return prediction

    @torch.no_grad()
    def _visualize_pred(self, img, mask, pred, im_iou, meta: Dict) -> wandb.Image:
        if pred.shape[-2:] != img.shape[-2:]:
            # Add channel dim, resize, remove channel dim:
            mask = F.resize(
                mask.unsqueeze(0),
                img.shape[-2:],
                interpolation=InterpolationMode.NEAREST,
            ).squeeze()
            pred = F.resize(
                pred.unsqueeze(0),
                img.shape[-2:],
                interpolation=InterpolationMode.NEAREST,
            ).squeeze()

        metastr = json.dumps(meta, default=str)
        masks = {}
        undefined_display_value = len(self.class_names) + 1

        if len(self.class_names) <= mtl_settings.max_classes_detailed_logging:
            for i, class_name in enumerate(self.class_names):
                # predictions
                pred_class = pred[i].clone()
                pred_class[pred_class == 1] = pred_class[pred_class == 1] * (i + 1)
                if mtl_settings.ignore_class_value is not None:
                    pred_class[pred_class == mtl_settings.ignore_class_value] = undefined_display_value
                # ground truth
                mask_class = mask[i].clone()
                mask_class[mask_class == 1] = mask_class[mask_class == 1] * (i + 1)
                if mtl_settings.ignore_class_value is not None:
                    mask_class[mask_class == mtl_settings.ignore_class_value] = undefined_display_value

                masks[f"predictions_{class_name}"] = {
                    "mask_data": pred_class.numpy(),
                    "class_labels": {
                        undefined_display_value: "ignored",
                        i + 1: class_name,
                    },
                }
                masks[f"ground_truth_{class_name}"] = {
                    "mask_data": mask_class.numpy(),
                    "class_labels": {
                        undefined_display_value: "ignored",
                        i + 1: class_name,
                    },
                }
        res = wandb.Image(
            img,
            masks=masks,
            caption=f"{self.class_names}\nIOU: {im_iou.numpy()}\n[{torch.min(img):.2f}, {torch.max(img):.2f}]\n{img.shape}\n{metastr}]",
        )
        return res

    def normalize_loss(self, dice_loss: torch.Tensor, focal_loss: torch.Tensor) -> torch.Tensor:
        return dice_loss + (focal_loss * 1.44269)  # log2(exp(1))

    # def log_epoch_metrics(self) -> Tuple[Dict[str, Any], str]:
    #     return {}, f"{self.get_name()} - loss: {self.get_short_status()}"



from typing import Any
from pathlib import Path
import numpy as np
from io import BytesIO
import requests
from PIL import Image
import imageio.v3 as imageio
import torchvision.transforms.functional as F
import torch
from sklearn.ensemble import RandomForestClassifier
import albumentations as A
from torchvision.transforms import InterpolationMode

# Enable optional dependencies
try:
    from label_studio_sdk import Client, Project
except ImportError:
    Client, Project = Any, Any

from mmm.interactive import pipes


@torch.inference_mode()
def img_to_segmentationfeatures(img: torch.Tensor, shape, encoder, decoder):
    inps = img.to(encoder.torch_device).unsqueeze(0)
    pyr, _ = encoder(inps)
    features = decoder(pyr)
    if shape != features.shape[1:]:
        features = F.resize(features, shape, interpolation=InterpolationMode.BILINEAR)
    return features.cpu()


def case_to_sklearn(case: dict[str, any], features: torch.Tensor, ignore_index=None):
    if ignore_index is None:
        # All rows and columns are relevant
        rows, cols = torch.where(torch.ones_like(case["label"]))
    else:
        rows, cols = torch.where(case["label"] != ignore_index)
    X = features[:, rows, cols].T
    y = case["label"][rows, cols]
    return X, y


def cases_to_sklearn(cases: list[dict[str, any]], features: torch.Tensor, ignore_index=0):
    cases = [case_to_sklearn(case, features[i], ignore_index) for i, case in enumerate(cases)]
    X = torch.cat([case[0] for case in cases])
    y = torch.cat([case[1] for case in cases])
    return X.numpy(), y.numpy()


if __name__ == "__main__":
    import pydantic

    pydantic.class_validators._FUNCS.clear()
    import streamlit as st
    from mmm.logging.st_ext import blend_with_mask

    LSURL = "http://localhost:8080"
    LSTOKEN = "2ac7ff7ae38466ebd19122ae019d11092176da2c"
    CLASS_NAMES = ["nonannotated", "background", "heart", "lung"]
    MODULES_PATH = Path("/host-home/shares/imageData/deep_learning/output/jschaefer/univ_sh_modules/")

    lsclient = Client(LSURL, LSTOKEN)
    project: Project = lsclient.get_project(lsclient.get_projects()[0].id)
    st.title(f"Project with {len(project.tasks)} tasks")

    preparer = pipes.Alb([A.Resize(224, 224)])

    # build cases from annotations
    cases = build_all_cases_for_project(project, transform=preparer)

    encoder = torch.load(MODULES_PATH / "encoder.pt")
    decoder = torch.load(MODULES_PATH / "decoder.pt")
    features = img_to_segmentationfeatures([case["image"] for case in cases], cases[0]["label"].shape, encoder, decoder)
    X, y = cases_to_sklearn(cases, features)

    rf = RandomForestClassifier().fit(X, y)

    fullX, _ = case_to_sklearn(cases[0], features[0], ignore_index=None)
    preds = rf.predict(fullX).reshape(cases[0]["label"].shape)
    st.write(fullX.shape)
    st.write(preds.shape)
    blend_with_mask(cases[0]["image"], torch.from_numpy(preds), classes=CLASS_NAMES, st_key="preds")



from typing import Literal, List, Mapping, Sequence, Union

import torch

from .SharedBlock import SharedBlock
from mmm.neural.modules.smp_modules import (
    SMPUnetDecoder,
    SMPPyramidAttentionDecoder,
)
from mmm.neural.modules.smp_modules import SMPPyramidAttentionDecoderConfig


class PyramidDecoder(SharedBlock):
    """
    Wraps a decoder from SMP.
    """

    class Config(SharedBlock.Config):
        module_name: str = "decoder"
        model: Union[SMPUnetDecoder.Config, SMPPyramidAttentionDecoderConfig] = SMPUnetDecoder.Config()

    def __init__(self, args: Config, enc_out_channels: List[int], encoder_output_stride: int) -> None:
        super().__init__(args)
        self.enc_out_channels, self.encoder_output_stride = (
            enc_out_channels,
            encoder_output_stride,
        )
        self.args: PyramidDecoder.Config
        self.model: Union[SMPUnetDecoder, SMPPyramidAttentionDecoder] = self.args.model.build_instance(
            enc_out_channels, encoder_output_stride
        )
        self.make_mtl_compatible()

    def get_output_dim_per_pixel(self) -> int:
        return self.model.get_output_dim_per_pixel()

    def get_upsampling_factor(self) -> int:
        """
        Used by segmentation task to know how much interpolation is required to reconstruct a mask with the input's size
        """
        return self.model.get_upsampling_factor()

    def forward(self, feature_pyramid: List[torch.Tensor]) -> torch.Tensor:
        return self.model(feature_pyramid)

    def get_example_input(self):
        # Usually, the feature maps would have sizes corresponding to the encoder output strides.
        example_input = [torch.rand(1, self.enc_out_channels[0], 224, 224)] + [
            torch.rand(1, c, 224 // (2 ** (i + 2)), 224 // (2 ** (i + 2))).to(self.torch_device)
            for i, c in enumerate(self.enc_out_channels[1:])
        ]
        # assert example_input[-1].shape[3] == 224 // self.encoder_output_stride
        return example_input

    def get_input_names(self) -> Sequence[str]:
        return ["input"] + [f"pyramid_{i}" for i in range(len(self.enc_out_channels) - 1)]

    def get_output_names(self) -> Sequence[str]:
        return ["pixel_embedding"]

    def get_dynamic_axes(self) -> Mapping[str, Mapping[int, str]]:
        return {
            "input": {0: "batch_size"},
            "pixel_embedding": {0: "batch_size", 2: "height", 3: "width"},
            **{
                f"pyramid_{i}": {
                    0: "batch_size",
                    2: f"height_lvl{i}",
                    3: f"width_lvl{i}",
                }
                for i in range(len(self.enc_out_channels) - 1)
            },
        }



import logging
from typing import Any, Literal, Tuple
from typing_extensions import Unpack
from pydantic import Field
from pydantic.config import ConfigDict
import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision.transforms import transforms
from torchvision.models import vgg16, VGG16_Weights, resnet18

from .TorchModule import TorchModule
import segmentation_models_pytorch.losses as smp_losses

from mmm.torch_ext import replace_childen_recursive
from mmm.BaseModel import BaseModel


class CrossEntropyLossConfig(TorchModule):
    loss_type: Literal["cross_entropy"] = "cross_entropy"

    def build_instance(self, *args, **kwargs) -> nn.Module:
        return nn.CrossEntropyLoss()


class NLLLossConfig(TorchModule):
    loss_type: Literal["negative_log_likelihood"] = "negative_log_likelihood"

    def build_instance(self, *args, **kwargs) -> nn.Module:
        return NLLLoss()


class NLLLoss(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        # NLLLoss excpets softmaxed pred. LogSoftmax used bcs of pytorch
        # https://pytorch.org/docs/stable/generated/torch.nn.Softmax.html
        self.loss_fc = nn.NLLLoss()
        self.soft = nn.LogSoftmax(dim=1)

    def __call__(self, y_pred, y_true):
        return self.loss_fc(self.soft(y_pred), y_true)


class MSELossConfig(TorchModule):
    loss_type: Literal["mean_squared_error"] = "mean_squared_error"
    value_range: Tuple[float, float] = Field(
        default=(0.0, 1.0),
        description="Expected range (min, max) of the values passed into the loss function. Used to scale the loss to [0, 1].",
    )

    def build_instance(self, *args, **kwargs) -> nn.Module:
        return MSELoss(self)


class RMSELossConfig(TorchModule):
    loss_type: Literal["root_mean_squared_error"] = "root_mean_squared_error"
    value_range: Tuple[float, float] = Field(
        default=(0.0, 1.0),
        description="Expected range (min, max) of the values passed into the loss function. Used to scale the loss to [0, 1].",
    )

    def build_instance(self, *args, **kwargs) -> nn.Module:
        return RMSELoss(self)


class FocalLossConfig(TorchModule):
    """
    Focal loss helps to focus your training on the difficult examples.
    """

    loss_type: Literal["focal"] = "focal"
    alpha: float = 1.0
    gamma: float = 2.0
    logits: bool = True
    reduce: bool = True

    def build_instance(self, *args, **kwargs) -> nn.Module:
        return FocalLoss(self)


class MSELoss(nn.Module):
    def __init__(self, args: MSELossConfig) -> None:
        super().__init__()
        self.mse_loss = nn.MSELoss()

        range_min, range_max = args.value_range
        self.factor = 1.0 / (range_max - range_min)

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor):
        return self.mse_loss(inputs * self.factor, targets * self.factor)


class RMSELoss(nn.Module):
    def __init__(self, args: RMSELossConfig) -> None:
        super().__init__()
        self.mse_loss = MSELoss(MSELossConfig(value_range=args.value_range))

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor):
        return torch.sqrt(self.mse_loss(inputs, targets))


class FocalLoss(nn.Module):
    """
    Loss based on cross entropy that emphasizes those outputs that have a large difference to the targets.
    Focal loss is a simple trick which can be used to train networks when class imbalance is present.
    Focusing parameter gamma: Increase to emphasize hard examples and put less effort into optimizing easy ones.

    For a drop in replacement of nn.CrossEntropyLoss, use the default values
    """

    def __init__(self, args: FocalLossConfig):
        super().__init__()
        self.args = args
        # self.alpha, self.gamma, self.logits, self.reduce = alpha, gamma, logits, reduce
        self.ce_loss = nn.CrossEntropyLoss(reduction="none")

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor):
        # targets = targets.long()
        if self.args.logits:
            bce_loss = self.ce_loss(inputs, targets)
        else:
            bce_loss = F.cross_entropy(inputs, targets, reduce=None)
        pt = torch.exp(-bce_loss)
        f_loss = self.args.alpha * (1 - pt) ** self.args.gamma * bce_loss

        if self.args.reduce:
            return torch.mean(f_loss)
        else:
            return f_loss


class SMPFocalLoss:
    class Config(BaseModel):
        gamma: float = 2.0

    def __init__(self, cfg: Config, mode: Literal["multiclass", "multilabel"], ignore_index) -> None:
        self.crit = smp_losses.FocalLoss(mode=mode, ignore_index=ignore_index, gamma=cfg.gamma)

    def __call__(self, preds: torch.Tensor, targets: torch.Tensor):
        l = self.crit(preds, targets)
        if l.isnan():
            logging.warning(f"Focal loss is nan for {preds.shape=} and {targets.shape=}")
            return 0
        else:
            return l


class SMPDiceFocalLoss2D:
    class Config(BaseModel):
        focal: SMPFocalLoss.Config = SMPFocalLoss.Config()

    def __init__(
        self,
        cfg: Config,
        ignore_index,
        mode: Literal["multiclass", "multilabel"] = "multiclass",
    ) -> None:
        self.crit1 = SMPFocalLoss(cfg.focal, mode=mode, ignore_index=ignore_index)
        self.crit2 = smp_losses.DiceLoss(mode=mode, ignore_index=ignore_index)

    def __call__(self, preds, mask):
        f_loss = self.crit1(preds, mask)
        # Dice loss is 0 for perfect prediction, 1 for worst possible prediction
        d_loss = self.crit2(preds, mask)
        return {"focal": f_loss, "dice": d_loss}



import logging
import numpy as np
import torch
import cv2
from typing import List, Optional, Callable, Dict, Any, TypeVar, Tuple

from torch.utils.data import Dataset
import torchvision.transforms as transforms
from .MTLDataset import MTLDataset, SrcCaseType
from .SemSegDataset import SemSegDataset

try:
    from mmdet.evaluation.functional import eval_map
except ImportError:
    eval_map = Any


def get_mmdet_annos(gtboxes, gtlabels):
    return [{"bboxes": box, "labels": lbl} for box, lbl in zip(gtboxes, gtlabels)]


def get_mmdet_preds(predboxes, predlabels, class_names: List[str]):
    det_results = []
    for image_pred_boxes, image_pred_labels in zip(predboxes, predlabels):
        # A list of predictions for each class
        class_boxes = [
            [box for i, box in enumerate(image_pred_boxes) if image_pred_labels[i] == cls_index]
            for cls_index in range(len(class_names))
        ]
        det_results.append([np.array(boxes) if boxes else np.empty((0, 5)) for boxes in class_boxes])
    return det_results


def eval_map_batch(predboxes, predlabels, gtboxes, gtlabels, class_names: List[str]):
    preds = get_mmdet_preds(predboxes, predlabels, class_names)
    annos = get_mmdet_annos(gtboxes, gtlabels)

    return eval_map(
        preds,
        annos,
        logger="silent",
    )


class DetectionDataset(MTLDataset):
    """
    A 2D detection dataset should consist of:

    - "image" -> FloatTensor[C, H, W] in the 0-1 range
    - "boxes" -> FloatTensor[N, 4] in [x1, y1, x2, y2] format (pascal_voc format)
    - "labels" -> Int64Tensor[N], the class label for each box
    """

    @staticmethod
    def convert_coco_box_to_pascal_box(x1, y1, w, h):
        return [max(x1, 0.0), max(y1, 0.0), x1 + w, y1 + h]

    def batch_collater(self) -> Callable:
        """
        For detection tasks the common batch format seems to be a list of cases.
        By default, collate_fn overwriters work with a list of cases.
        """
        # No inputs are stacked, detection tasks want lists!
        batch_type = TypeVar("batch_type", bound=List[Dict[str, Any]])

        def f(x: batch_type) -> batch_type:
            return x

        return f

    @staticmethod
    def from_semseg(ds: SemSegDataset, for_class_indices: Tuple[int, ...]):
        """Computes connected components in the mask and each connected component will get a box"""

        def convert_semseg_to_detection_case(semseg_case: Dict[str, Any]) -> Dict[str, Any]:
            assert "image" in semseg_case and "label" in semseg_case

            boxes = []
            box_labels = []

            for class_index in for_class_indices:
                lbl_for_class = (semseg_case["label"].numpy() == class_index).astype(np.uint8)
                contours, _ = cv2.findContours(lbl_for_class, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
                for c in contours:
                    x, y, w, h = cv2.boundingRect(c)
                    # Suppress boxes from a single pixel
                    # if w > 0 and h > 0:
                    boxes.append([x, y, (x + w), y + h])
                    box_labels.append(class_index)

            semseg_case["boxes"] = torch.FloatTensor(boxes)
            semseg_case["labels"] = torch.LongTensor(box_labels)

            return semseg_case

        return DetectionDataset(
            src_ds=ds,
            src_transform=convert_semseg_to_detection_case,
            class_names=ds.class_names,
        )

    def __init__(
        self,
        src_ds: Dataset[SrcCaseType],
        src_transform: Optional[Callable[[SrcCaseType], Dict[str, Any]]] = None,
        class_names: Optional[List[str]] = None,
        collate_fn: Optional[Callable] = None,
        *args,
        **kwargs,
    ) -> None:
        assert class_names is not None, "Detection datasets without class names are deprecated"
        self.vis_classes: List[str] = class_names
        if collate_fn is not None:
            collate_fn = transforms.Compose([collate_fn, self.batch_collater()])
        else:
            collate_fn = self.batch_collater()
        super().__init__(
            src_ds,
            ["image", "boxes", "labels"],
            ["original_size", "meta"],
            src_transform,
            collate_fn=collate_fn,
            *args,
            **kwargs,
        )

    def verify_case_by_index(self, index: int) -> Dict[str, Any]:
        case = super().verify_case_by_index(index)
        self.assert_image_data_assumptions(case["image"])
        assert (
            case["boxes"].shape[0] == case["labels"].shape[0]
        ), f"Number of boxes and box labels do not match: {case['boxes'].shape[0]=} {case['labels'].shape[0]=}"
        # if len(case['boxes'].shape) != 2:
        #     print("test")
        if case["boxes"].shape[0] > 0:  # there are boxes
            assert len(case["boxes"].shape) == 2, f"{len(case['boxes'].shape)=} should be 2"
            assert case["boxes"].shape[-1] == 4, f"{case['boxes'].shape=}"

            assert isinstance(case["boxes"], torch.FloatTensor)
            for box in case["boxes"]:
                x1, y1, x2, y2 = box
                height, width = case["image"].shape[1:]
                conds = [
                    x1 >= 0,
                    x1 <= x2,
                    x2 >= x1,
                    x2 <= width,
                    y1 >= 0,
                    y1 <= y2,
                    y2 >= y1,
                    y2 <= height,
                ]
                assert False not in conds, f"Box sanity check failed for {box} because {conds=}"

            assert isinstance(case["labels"], torch.LongTensor)
            if self.vis_classes is not None:
                # Make sure the labels make sense
                for label in case["labels"]:
                    assert label >= 0 <= len(self.vis_classes)

        return case

    def set_classes_for_visualization(self, classes: List[str]):
        self.vis_classes = classes

    def get_classes_for_visualization(
        self,
    ):
        assert self.vis_classes is not None
        return self.vis_classes

    def st_case_viewer(self, case: Dict[str, Any], i: int) -> None:
        import streamlit as st
        from mmm.logging.st_ext import blend_with_mask

        st.title("Untransformed image:")
        im = case["image"]
        blend_with_mask(
            im,
            None,
            caption_suffix=f"Shape: {im.shape}",
            classes=self.vis_classes,
            boxes=(case["boxes"], case["labels"]),
            st_key=f"c{i}",
        )
        st.write(case)

    def _visualize_batch_case(self, batch: List[Dict[str, Any]], i: int) -> None:
        import streamlit as st
        from mmm.logging.st_ext import blend_with_mask

        patch = batch[i]["image"]
        patch_boxes = batch[i]["boxes"]
        patch_labels = batch[i]["labels"]
        # class_name = "No class names known" if self.vis_classes is None else self.vis_classes[batch['class'][i]]
        # st.write(f"Label: {batch['class'][i]}, " + class_name)
        blend_with_mask(
            patch,
            None,
            caption_suffix=f"{i}/{self._compute_batchsize_from_batch(batch)}: {patch.shape}",
            classes=self.vis_classes,
            boxes=(patch_boxes, patch_labels),
            st_key=f"b{i}",
        )
        if "meta" in batch[i]:
            st.write(batch[i]["meta"])

    def _compute_batchsize_from_batch(self, batch: List[Dict[str, Any]]) -> int:
        return len(batch)



from typing import Mapping, Tuple
from pydantic import Field

import torch
import torch.nn as nn

from mmm.mtl_modules.shared_blocks.SharedBlock import ModelInput, SharedBlock
from mmm.neural.activations import ActivationFn, ActivationFunctionConfig


class Squeezer(SharedBlock):
    """
    Takes a feature pyramid and squeezes it into a single feature map.
    """

    class Config(SharedBlock.Config):
        module_name: str = "squeezer"
        out_channels: int = Field(
            -1,
            description="""Number of output channels, -1 for number of input channels.
            In that case, no convolution is applied.
            """,
        )
        activation: ActivationFunctionConfig = ActivationFunctionConfig(fn_type=ActivationFn.GeLU)
        use_channel_attention: bool = False

    def __init__(self, args: Config, enc_out_channels: list[int], enc_strides: list[int]):
        super().__init__(args)
        self.args: Squeezer.Config
        self.enc_out_channels = enc_out_channels
        self.enc_strides = enc_strides

        # self.norm = nn.GroupNorm(num_groups=1, num_channels=self.get_hidden_dim())
        # Use batchnorm because it will be automatically converted into the correct type of norm
        self.norm = nn.BatchNorm2d(num_features=self.get_hidden_dim())
        self.conv = (
            nn.Conv2d(
                in_channels=enc_out_channels[-1],
                out_channels=self.get_hidden_dim(),
                kernel_size=1,
                stride=1,
                padding=0,
            )
            if self.args.out_channels != -1
            else nn.Identity()
        )
        self.activation = args.activation.build_instance()
        self.avg = nn.AdaptiveAvgPool2d((1, 1))
        if self.args.use_channel_attention:
            self.m = nn.AdaptiveMaxPool2d((1, 1))
        self.make_mtl_compatible()

    def get_hidden_dim(self) -> int:
        return self.args.out_channels if self.args.out_channels > 0 else self.enc_out_channels[-1]

    def forward(self, feature_pyramid: list[torch.Tensor]) -> tuple[torch.Tensor, torch.Tensor]:
        # The order is inspired by torchvision.ops.misc.ConvNormActivation
        # experimental setting to increase the influence between pyramidal and pooling tasks
        if self.args.use_channel_attention:
            atten = self.avg(feature_pyramid[-1]) + self.m(feature_pyramid[-1])
            feat_latent = self.norm(self.conv(atten))
            feat = self.norm(self.conv(feature_pyramid[-1]))
            return self.activation(feat + feat_latent), self.activation(feat_latent)
        else:
            return self.activation(self.norm(self.conv(feature_pyramid[-1]))), self.activation(
                self.norm(self.conv(self.avg(feature_pyramid[-1])))
            )

    def get_example_input(self) -> ModelInput | Tuple[ModelInput, ...]:
        return [torch.rand(1, self.enc_out_channels[-1], 7, 7).to(self.torch_device)]

    def get_dynamic_axes(self) -> Mapping[str, Mapping[int, str]]:
        return {"input": {0: "batch_size", 2: "height", 3: "width"}}



from __future__ import annotations
from typing import Any, List, Optional, Callable, Dict, Tuple, TypeVar
from functools import partial

import torch
import torchvision.transforms.functional as F
import torchvision.transforms as transforms
from torch.utils.data import Dataset

from .MTLDataset import MTLDataset, SrcCaseType
from .SemSegDataset import SemSegDataset
from mmm.transforms import KeepOnlyKeysInDict


class MultilabelClassificationDataset(MTLDataset):
    """
    Expects cases with the keywords:

    - image: (typical image assumptions of mmm)
    - class_labels: torch.FloatTensor with a float between 0. and 1. for the confidence of that class being True

    Optional:

    - loss_weights: Allows to ignore certain class during loss computation. Use this to deal e.g. with NaNs targets
    """

    @staticmethod
    def from_semseg(ds: SemSegDataset, ignore_classes: Optional[List[int]] = None) -> MultilabelClassificationDataset:
        """
        Uses ds in-place to construct a multilabel classification dataset by appending a batch transform.
        """
        if ignore_classes is None:
            ignore_classes = []
        assert len(ignore_classes) == len(set(ignore_classes)), f"Don't ignore a class twice {ignore_classes=}"

        def semsegbatch_to_mclf_batch(num_classes: int, c, semsegbatch: Dict[str, Any]) -> Dict[str, Any]:
            masks = semsegbatch["label"]
            class_labels = []
            for i in range(masks.shape[0]):
                x = torch.zeros(num_classes)
                original_classes = torch.unique(masks[i, :]).tolist()
                x[[c[x] for x in original_classes if x in c]] = 1
                class_labels.append(x)

            semsegbatch["class_labels"] = torch.stack(class_labels)
            return semsegbatch

        class_names = [v for i, v in enumerate(ds.class_names) if i not in ignore_classes]
        conv = {}
        for i, _ in enumerate(ds.class_names):
            if i not in ignore_classes:
                conv[i] = len(conv)

        mclf_ds = MultilabelClassificationDataset(
            ds.src_ds,
            class_names=class_names,
            src_transform=ds.src_transform,
            batch_transform=ds.batch_transform,
            collate_fn=transforms.Compose(
                [
                    ds.collate_fn,
                    partial(semsegbatch_to_mclf_batch, len(class_names), conv),
                ]
            ),
        )
        # For this approach to work the data stripper needs to chill and allow semseg and multilabel keys
        mclf_ds.data_stripper = KeepOnlyKeysInDict(
            keys=set(list(ds.data_stripper.keys) + list(mclf_ds.data_stripper.keys)),
        )

        return mclf_ds

    def __init__(self, src_ds: Dataset[SrcCaseType], class_names: List[str], **kwargs) -> None:
        self.class_names: List[str] = class_names
        super().__init__(src_ds, ["image", "class_labels"], ["loss_weights", "meta"], **kwargs)

    def verify_case_by_index(self, index: int) -> Dict[str, Any]:
        case = super().verify_case_by_index(index)
        self.assert_image_data_assumptions(case["image"])
        assert isinstance(case["class_labels"], torch.FloatTensor), "Labels should be confidences between 0. and 1."
        assert len(case["class_labels"]) == len(self.class_names)
        assert torch.min(case["class_labels"]) >= 0.0 and torch.max(case["class_labels"]) <= 1.0

        if "loss_weights" in case:
            assert isinstance(case["loss_weights"], torch.FloatTensor), "Loss weights should be float between 0. and 1."
            assert len(case["loss_weights"]) == len(self.class_names)
            assert torch.min(case["loss_weights"]) >= 0.0 and torch.max(case["loss_weights"]) <= 1.0
        return case

    def get_input_output_tuple(self, batch: Dict[str, Any]) -> Tuple[Any, ...]:
        return batch["image"], batch["class_labels"]

    def st_case_viewer(self, case: Dict[str, Any], i: int) -> None:
        import streamlit as st
        from mmm.logging.st_ext import blend_with_mask

        st.title("Untransformed image:")
        im = case["image"]
        blend_with_mask(im, None, caption_suffix=f"Shape: {im.shape}", st_key=f"c{i}")
        self._print_relevant_classes(
            case["class_labels"],
        )
        st.write(case)

    def _compute_batchsize_from_batch(self, batch: Dict[str, Any]) -> int:
        return batch["image"].shape[0]

    def _visualize_batch_case(self, batch: Dict[str, Any], i: int) -> None:
        import streamlit as st
        from mmm.logging.st_ext import blend_with_mask

        patch, class_labels = batch["image"][i], batch["class_labels"][i]

        self._print_relevant_classes(class_labels, batch["loss_weights"][i] if "loss_weights" in batch else None)
        if "meta" in batch:
            st.json(batch["meta"][i])
        blend_with_mask(
            patch,
            None,
            caption_suffix=f"{i}/{self._compute_batchsize_from_batch(batch)}: {patch.shape}",
            st_key=f"b{i}",
        )

    def _print_relevant_classes(self, class_labels: torch.Tensor, loss_weights: Optional[torch.Tensor] = None):
        import streamlit as st

        ignored_classes = []
        st.write(f"{class_labels} (Labels)")
        if loss_weights is not None:
            st.write(f"{loss_weights} (Loss weights)")
        for i, v in enumerate(class_labels):
            if loss_weights is None or loss_weights[i] > 0:
                if v > 0.0:
                    st.success(f"{self.class_names[i]} ({i=}) -> {v}")
                else:
                    st.error(f"{self.class_names[i]} ({i=}) -> {v}")
            else:
                ignored_classes.append((self.class_names[i], i))

        st.write(f"Ignored classes: {ignored_classes}")



from functools import partial
from pydantic import Field
from typing import Dict, List, Literal, Optional
import torch
import torch.nn as nn
from torchvision.models import swin_v2_t, swin_v2_s, swin_v2_b
from torchvision.models.swin_transformer import (
    Swin_V2_T_Weights,
    Swin_V2_B_Weights,
    Swin_V2_S_Weights,
    SwinTransformer,
)
from ..TorchModule import TorchModule
from ..activations import ActivationFunctionConfig, ActivationFn
from mmm.torch_ext import infer_stride_channels_from_features

_models = {"tiny": swin_v2_t, "small": swin_v2_s, "base": swin_v2_b}

_weights = {
    "tiny": Swin_V2_T_Weights.DEFAULT,
    "small": Swin_V2_S_Weights.DEFAULT,
    "base": Swin_V2_B_Weights.DEFAULT,
}


class TorchVisionSwinformer(nn.Module):
    class Config(TorchModule):
        architecture: Literal["swinformer"] = "swinformer"
        pretrained: bool = False
        variant: Literal["tiny", "small", "base"] = "tiny"

        def build_instance(self, *args, **kwargs) -> nn.Module:
            return TorchVisionSwinformer(self)

    def __init__(self, args: Config) -> None:
        super().__init__()
        self.args = args
        # First load the full model to allow torchvision pretrained weights
        if args.pretrained:
            self.wrapped_model: SwinTransformer = _models[args.variant](weights=_weights[args.variant])  # type: ignore
        else:
            self.wrapped_model: SwinTransformer = _models[args.variant]()

        # We only need the features, not the head
        del self.wrapped_model.head

        with torch.no_grad():
            self.channels, self.strides = infer_stride_channels_from_features(
                self._feature_forward(torch.rand(1, 3, 224, 224))
            )
            # Currently strides are assumed to always be:
            assert self.strides == [1, 4, 8, 16, 32]

    def _feature_forward(self, x):
        feature_maps = [x]
        # self.wrapped_model.features are the layers which compute features in the swin transformer
        for l in self.wrapped_model.features:
            feature_maps.append(l(feature_maps[-1]))

        feature_maps = [x] + [feature_map.permute(0, 3, 1, 2) for feature_map in feature_maps[1:]]
        return feature_maps[::2]

    def get_strides(self):
        # Input and spatial sizes of the feature maps
        return self.strides

    def forward(self, x):
        pyr = self._feature_forward(x)

        return pyr

    def get_feature_pyramid_channels(self) -> List[int]:
        return self.channels



from __future__ import annotations

import logging
import json
from enum import Enum
from abc import abstractmethod
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Literal,
    Union,
)
from typing_extensions import Annotated
from datetime import datetime
from mmm.BaseModel import BaseModel
from pydantic import Field
from tqdm.auto import tqdm

import torch
from torch.cuda import OutOfMemoryError
import time
import wandb

from mmm.mtl_modules.MTLModule import MTLModule
from mmm.optimization.MTLOptimizer import MTLOptimizer
from mmm.task_sampling import (
    BaseSampler,
    TaskSamplerTypes,
    TaskSamplerConfig,
    CyclicTaskSampler,
    BalancedTaskSampler,
)
from mmm.mtl_modules.tasks.MTLTask import MTLTask
from mmm.mtl_modules.shared_blocks.SharedBlock import SharedBlock
from mmm.mtl_modules.shared_blocks.SharedModules import SharedModules
from mmm.DataSplit import DataSplit
from mmm.logging.type_ext import StepFeedbackDict
import torch.distributed as dist


class FixedMultistep(BaseModel):
    """
    Does an optimization step after a fixed number of forward passes `step_num`.
    """

    multistep_type: Literal["fixed"] = "fixed"
    step_num: int = 1


class LinearMultistep(BaseModel):
    """
    Scales the number of accumulated steps with the number of tasks.

    For example, if you use three tasks and factor one there will be three forward passes per optimization step.
    """

    multistep_type: Literal["linear"] = "linear"
    factor: int = 1


MultistepMode = Union[FixedMultistep, LinearMultistep]


class LoopLogConfig(BaseModel):
    # log_gradients_mode: Literal['every_500', 'every_update_step', 'once_per_epoch'] = 'every_500'
    print_dateformat: str = "%H:%M.%S"
    log_result: bool = Field(
        default=True,
        description="If true, logs more detailed statistics after the loop",
    )
    include_mem: bool = False


class LoopConfig(BaseModel):
    max_steps: int = Field(default=-1, description="Maximum steps per loop, -1 for unlimited")
    log_args: LoopLogConfig = LoopLogConfig()
    multistep_mode: Annotated[MultistepMode, Field(discriminator="multistep_type")] = LinearMultistep()

    task_sampler: Annotated[TaskSamplerConfig, Field(discriminator="sampler_type")]

    def _compute_multistep_num(self, num_tasks: int) -> int:
        if self.multistep_mode.multistep_type == FixedMultistep().multistep_type:
            return self.multistep_mode.step_num
        elif self.multistep_mode.multistep_type == LinearMultistep().multistep_type:
            return self.multistep_mode.factor * num_tasks
        else:
            raise Exception(f"Unknown multistep mode {self.multistep_mode}")

    @abstractmethod
    def build_instance(
        self,
        tasks: List[MTLTask],
        epoch: int,
        task_step: Callable[[Any, MTLTask], Tuple[torch.FloatTensor, StepFeedbackDict]],
        shared_blocks: List[SharedBlock],
        prefix: str,
        step_counters: Dict[str, int],
        optim: MTLOptimizer,
    ) -> Loop:
        raise NotImplementedError


class TrainLoopConfig(LoopConfig):
    task_sampler: Annotated[TaskSamplerConfig, Field(discriminator="sampler_type")] = CyclicTaskSampler.Config(
        mode="break_with_longest_loader"
    )

    def build_instance(self, tasks: List[MTLTask], *args, **kwargs) -> Loop:
        task_sampler = TaskSamplerTypes[self.task_sampler.sampler_type](self.task_sampler, tasks, DataSplit.train)
        multistep_num = self._compute_multistep_num(len(tasks))

        return Loop(self, task_sampler, multistep_num, True, *args, **kwargs)


class ValLoopConfig(LoopConfig):
    task_sampler: Annotated[TaskSamplerConfig, Field(discriminator="sampler_type")] = BalancedTaskSampler.Config()
    multistep_mode: Annotated[MultistepMode, Field(discriminator="multistep_type")] = FixedMultistep(step_num=1)

    def build_instance(self, tasks: List[MTLTask], *args, **kwargs) -> Loop:
        task_sampler = TaskSamplerTypes[self.task_sampler.sampler_type](self.task_sampler, tasks, DataSplit.val)
        # multistep_num = self._compute_multistep_num(len(tasks))

        return Loop(self, task_sampler, -1, False, *args, **kwargs)


class Loop:
    """
    Loops through a tasksampler.

    - prefix is the name of the loop. Controls the step which is logged.
    """

    def __init__(
        self,
        args: LoopConfig,
        task_sampler: BaseSampler,
        accumulate_losses_for_steps: int,
        training_mode: bool,
        epoch: int,
        task_step: Callable,
        shared_blocks: SharedModules,
        prefix: str,
        step_counters: Dict[str, int],
        optim: Optional[MTLOptimizer] = None,
        rank: int = 0,
        syncon_syncoff: Optional[Tuple[Callable, Callable]] = None,
    ) -> None:
        self.args: LoopConfig = args
        self.epoch: int = epoch
        self.task_step = task_step
        self.training_mode: bool = training_mode
        self.task_sampler: BaseSampler = task_sampler
        self.shared_blocks = shared_blocks
        self.prefix: str = prefix
        self.step_counters: Dict[str, int] = step_counters
        self.optim = optim
        self.accumulate_losses_for_steps = accumulate_losses_for_steps
        self.rank = rank
        self.one_update_per_task = self.accumulate_losses_for_steps == len(self.task_sampler.tasks)
        if syncon_syncoff is None:
            self.sync_on, self.sync_off = None, None
        else:
            self.sync_on, self.sync_off = syncon_syncoff

        if self.args.log_args.log_result:
            printout = f"Loop {self.prefix}, rank {self.rank}: {[t.get_name() for t in self.task_sampler.tasks]} "
            if training_mode:
                printout += f"Accu.-steps: {self.accumulate_losses_for_steps}. "
                printout += f"Tasks reduced each step: {self.one_update_per_task=}"
            logging.info(printout)

    def drain(self) -> List[float]:
        return [x for _, x in self]

    def drain_to_dict(self) -> Dict[MTLTask, List[float]]:
        res: Dict[MTLTask, List[float]] = {t: [] for t in self.task_sampler.tasks}
        for task, task_loss in self:
            res[task].append(task_loss)
        return res

    def has_len(self) -> bool:
        if self.args.max_steps > 0:
            return True
        else:
            return self.task_sampler.is_finite()

    def __len__(self) -> int:
        if self.task_sampler.is_finite():
            return (
                min(self.args.max_steps, len(self.task_sampler)) if self.args.max_steps > 0 else len(self.task_sampler)
            )
        else:
            assert self.args.max_steps > 0, "Using infinite task samplers require a max number of steps per loop."
            return self.args.max_steps

    def _prepare_loop(self):
        # Make sure the user did not forget to make their model MTL compatible
        for shared_block in self.shared_blocks.module.shared_modules.values():  # type: ignore
            shared_block: SharedBlock
            shared_block.prepare_epoch(self.epoch, self.prefix, training_mode=self.training_mode)
            assert shared_block._made_mtl_compatible, f"{shared_block} not yet MTL compatible"

        frozen_blocks = [
            f"{m.get_name()}: {m.training=}"  # type: ignore
            for m in self.shared_blocks.module.shared_modules.values()
            if not m.training == self.training_mode
        ]
        if frozen_blocks:
            logging.warning(f"Assuming frozen blocks, because loop is in {self.training_mode=}: {frozen_blocks}")

        for task in self.task_sampler.tasks:
            assert task is not None
            task.prepare_epoch(self.epoch, self.prefix, training_mode=self.training_mode)
            assert (
                not task.is_currently_being_trained()
            ), f"Task {task.get_name()} thinks it is already being trained before the loop"

        if self.prefix not in self.step_counters:
            self.step_counters[self.prefix] = 0

        if self.training_mode:
            # Stuff like optimizer.zero_grad
            self.optim.reset_shared_grads()

        if dist.is_initialized() and self.sync_off is not None:
            self.sync_off()

    def _finish_up(self):
        logging.info(
            f"{datetime.now():{self.args.log_args.print_dateformat}}:"
            + f" {self.prefix} done with: {self._get_progbar_desc()}"
        )
        log_dict = {f"counter/{self.prefix}epoch": self.epoch}
        for task in self.task_sampler.tasks:
            task_res_dict, task_print_str = task.log_epoch_metrics()
            logging.info(task_print_str)
            for metric_name, metric_val in task_res_dict.items():
                log_dict[f"{task.get_name()}/{self.prefix}epoch.{metric_name}"] = metric_val

        if self.training_mode:
            optim_state: Dict[str, Any] = self.optim.get_current_logstate()
            for k, v in optim_state.items():
                log_dict[f"optimizer/{self.prefix}_{k}"] = v

        if self.args.log_args.log_result:
            wandb.log(log_dict)

    def _get_progbar_desc(self) -> str:
        task_statuses = [t.get_short_status() for t in self.task_sampler.tasks]
        return f"p{self.rank}_{self.prefix}_e{self.epoch}: {', '.join(task_statuses)}"

    def __iter__(self):
        self._prepare_loop()

        iteration_start_time, is_update_step = time.perf_counter(), False

        # assert self.task_sampler.is_finite() or self.args.max_steps > 0, \
        #     f"Either use a finite task sampler or specify the maximum number of steps in {self}"
        task_iterator: Iterator[Tuple[Any, MTLTask]] = self.task_sampler.continue_iter()

        def do_update_step() -> None:
            self.optim.after_iteration_step()
            self.optim.reset_shared_grads()

        with torch.set_grad_enabled(self.training_mode):
            with tqdm(
                total=self.__len__() if self.has_len() else None,
                leave=self.args.log_args.log_result,
                position=self.rank,
            ) as pbar:
                for i, (batch, task) in enumerate(task_iterator):
                    batch_extraction_ms = time.perf_counter() - iteration_start_time
                    before_mem = torch.cuda.memory_allocated()

                    # is_last_step = i >= self.__len__() - 1
                    is_update_step = (self.has_len() and (i >= self.__len__() - 1)) or (
                        i % self.accumulate_losses_for_steps == 0
                    )
                    is_update_step = self.training_mode and is_update_step and (i > 0)

                    assert task.training == self.training_mode
                    # Set active task for all shared blocks
                    for block in self.shared_blocks.module.shared_modules.values():  # type: ignore
                        block: SharedBlock
                        block.set_active_task(task.get_name())

                    if self.training_mode and is_update_step and (self.sync_on is not None):
                        # the forward step needs to know if a sync will happen, so turn on sync here already
                        self.sync_on()
                    # with Join([self.shared_blocks], enable=False):
                    try:
                        task_step_result = self.task_step(batch, task)
                    except RuntimeError as e:
                        if "out of memory" in str(e):
                            logging.error(
                                f"Task {task.get_name()} on rank {self.rank} encountered {e}, reduce size of {batch}!"
                            )
                            raise e
                        else:
                            torch.save(batch, f"batch_{self.rank}.pt")
                            logging.error(
                                f"Encountered runtimeerror {e} with task {task.get_name()} on rank {self.rank}!"
                            )
                            raise e

                    if task_step_result is not None:
                        task_step_loss, task_log_dict = task_step_result

                        if torch.isnan(task_step_loss).item():
                            raise Exception(f"Got a nan loss for \n{batch}\n in task: \n {task}")

                        before_backward_mem = torch.cuda.memory_allocated()
                        if self.training_mode:
                            task_step_loss.backward()
                            # Updates the task head if there is exactly one update per task in each iteration
                            self.optim.after_backward(
                                of_task=task,
                                reduce_task_immediately=self.one_update_per_task,
                            )
                            if self.sync_off is not None:
                                self.sync_off()
                        after_backward_mem = torch.cuda.memory_allocated()

                        if is_update_step:
                            # print("Finding unused parameters:")
                            # for name, param in self.shared_blocks.named_parameters():
                            #     if param.grad is None:
                            #         print(name)
                            do_update_step()

                        network_passes_ms = time.perf_counter() - iteration_start_time - batch_extraction_ms

                        step_prefix = f"{self.prefix}step"
                        task_uniqueness = f"_{task.get_name()}"
                        step_log_dict = {
                            f"counter/{self.prefix}epoch": self.epoch,
                            f"counter/{step_prefix}": self.step_counters[self.prefix],
                            f"{step_prefix}loss/batch_loss{task_uniqueness}": task_step_loss.item(),
                        }

                        if self.args.log_args.log_result:
                            task_log_dict = {
                                f"{task.get_name()}/{step_prefix}.{k}": v for k, v in task_log_dict.items()
                            }
                            step_log_dict.update(task_log_dict)

                        step_log_dict.update(
                            {
                                f"{step_prefix}time/batchextraction_time{task_uniqueness}": batch_extraction_ms,
                                f"{step_prefix}time/networkpasses_time{task_uniqueness}": network_passes_ms,
                                f"{step_prefix}time/fulliteration_time{task_uniqueness}": network_passes_ms
                                + batch_extraction_ms,
                            }
                        )

                        if self.args.log_args.include_mem:
                            mem_pref = f"{step_prefix}memory/"
                            step_log_dict[f"{mem_pref}afterstep{task_uniqueness}"] = torch.cuda.memory_allocated() / (
                                1024**3
                            )
                            step_log_dict[f"{mem_pref}beforestep{task_uniqueness}"] = before_mem / (1024**3)
                            step_log_dict[f"{mem_pref}beforeback{task_uniqueness}"] = before_backward_mem / (1024**3)
                            step_log_dict[f"{mem_pref}afterback{task_uniqueness}"] = after_backward_mem / (1024**3)

                        wandb.log(step_log_dict)

                        yield task, task_step_loss.item()
                    else:
                        logging.warn(f"Task {task.get_name()} wants to skip step {i}")

                    self.step_counters[self.prefix] += 1
                    iteration_start_time = time.perf_counter()
                    pbar.update(1)
                    pbar.set_description(self._get_progbar_desc())

                    if self.has_len() and (i >= self.__len__() - 1):
                        break

            if (not is_update_step) and self.training_mode:
                assert self.sync_off is None and self.sync_off is None, "Multi GPU not compatible with sudden loop end"
                # If the last step in the loop was not an update step there is an unused gradient
                do_update_step()

        self._finish_up()

    def __repr__(self) -> str:
        return f"""Loop with train: {self.training_mode}\n
        With Args:\n{json.dumps(self.args.dict(), indent=2)}\n
        For tasks: {[t.get_name() for t in self.task_sampler.tasks]}"""



"""
Utilities that extend the PyTorch types, not relying on our MTL extensions such as the MTLDataset
"""

from __future__ import annotations
import os
import logging
import random
from typing import (
    Callable,
    Generic,
    Optional,
    Sized,
    Union,
    TypeVar,
    Dict,
    cast,
    Iterator,
    List,
    Tuple,
)
from pathlib import Path
import numpy as np
import math
import torch
import torch.nn as nn
from torch.utils.data import (
    Dataset,
    ChainDataset,
    Subset,
    IterableDataset,
    get_worker_info,
    DataLoader,
)
from tqdm.auto import tqdm

from mmm.BaseModel import BaseModel
from mmm.data_loading.MTLDataset import MTLDataset
from mmm.utils import get_default_cachepath

T = TypeVar("T")


def replace_childen_recursive(m: nn.Module, layertype_to_replace, newlayer_constructor):
    """
    newlayer_constructor gets an instance of the layer to be replaced.

    Use cases might be replacing all 2D layers by their respective 3D versions.
    """
    for k, layer in m.named_children():
        # Replace the object in question
        if isinstance(layer, layertype_to_replace):
            newlayer = newlayer_constructor(layer)
            # Does not work for children's children, which is why this is a recursive function
            setattr(m, k, newlayer)

        replace_childen_recursive(layer, layertype_to_replace, newlayer_constructor)


def get_random_ds_sample(ds: Dataset[T], subset_size: Union[float, int]) -> Dataset[T]:
    """
    Generates a random subset of a given dataset.

    On the new object, only the default methods will work.
    """
    original_dataset_len = len(ds)  # type: ignore

    if isinstance(subset_size, float):
        subset_size = math.floor(original_dataset_len * subset_size)

    new_indices = random.sample(range(original_dataset_len), subset_size)

    return Subset(ds, new_indices)


def default_replacer(cache_size: int, cache: list[T], replacement_items: list[T]) -> tuple[list[T], list[T]]:
    cache.extend(replacement_items)
    # Only keep the last `cache_size` items
    cache = cache[-cache_size:]
    replacement_items = []
    return cache, replacement_items


class RunningCacheDataset(Dataset):
    """
    Uses a background worker to prepare items for the cache.

    It is intended to call `replace_cache_items()` every epoch, which will replace the oldest items in the cache
    with new items from the background loader.
    Alternatively, auto replacement may be used,
    which calls replacement after the number of items currently in the cache were yielded.

    For customizing the replacement of cache items with new items, use the replacer argument.
    For customizing the items that the background loader requests, use a sampler with the dataloader_kwargs.
    """

    class Config(BaseModel):
        cache_size: int = 1000
        background_loader_prefetch: int = 10
        background_workers: int = 1
        min_cache_size: int = 1

    def __init__(self, wrapped_dataset: Dataset, args: Config, replacer=None, **dataloader_kwargs):
        self.args = args
        self.cache = []
        self.replacement_items = []
        self.replacer = replacer if replacer is not None else default_replacer
        self.wrapped_ds, self.backgroundloader_kwargs = (
            wrapped_dataset,
            dataloader_kwargs,
        )
        self.background_loader = DataLoader(
            self.wrapped_ds,
            batch_size=1,
            num_workers=self.args.background_workers,
            prefetch_factor=self.args.background_loader_prefetch,
            # persistent_workers=True,
            collate_fn=lambda x: x[0],
            **self.backgroundloader_kwargs,
        )
        self.mp_iter = self.rebuild_backgrounditer()

        for _ in tqdm(range(self.args.min_cache_size)):
            self.cache.append(self.request_cache_item())

    def rebuild_backgrounditer(self):
        assert len(self.wrapped_ds) == len(self.background_loader)
        return iter(self.background_loader)

    def request_cache_item(self):
        try:
            return next(self.mp_iter)
        except StopIteration:
            # This restarts the background loader in case of iterable-style datasets
            self.mp_iter = self.rebuild_backgrounditer()
            return self.request_cache_item()

    def worker_has_data_available(self) -> bool:
        # self.mp_iter._data_queue is a multiprocessing.queues.Queue
        return self.mp_iter._data_queue.qsize() > 0

    def get_cache_max_size(self) -> int:
        try:
            ds_len = len(self.wrapped_ds)
        except:
            return self.args.cache_size
        return min(ds_len, self.args.cache_size)

    def __len__(self):
        return len(self.cache)

    def replace_cache_items(self):
        """Replaces cache items with new items from the background loader."""
        self.cache, self.replacement_items = self.replacer(
            self.get_cache_max_size(), self.cache, self.replacement_items
        )

    def __getitem__(self, idx):
        while self.worker_has_data_available() and len(self.replacement_items) < self.args.cache_size:
            self.replacement_items.append(self.request_cache_item())

        # If there is nothing to fetch anymore, we can restart the background loader
        # This restarts the background loader in case of map-style datasets
        if self.mp_iter._tasks_outstanding <= 0:
            assert self.mp_iter._data_queue.qsize() == 0, "There should be no data left in the queue"
            self.mp_iter = self.rebuild_backgrounditer()

        res = self.cache[idx]
        return res


SuperCaseType = TypeVar("SuperCaseType")
SubCaseType = TypeVar("SubCaseType")


class CachingSubCaseDS(IterableDataset, Generic[SubCaseType]):
    """
    Holds `cache_size` subcases in a cache for each worker.
    The cache is refilled with subcases once a new supercase fits into cache.

    Each supercase is assigned one worker, try to keep the number of workers low.
    In consequence, there is a sampling bias.
    For example, if you have 12 supercases and 4 workers, each worker can construct batches from at most 3 supercases.

    At the end of the loop all subcases are drained, resulting in a loading time at the start of each epoch.
    """

    class Config(BaseModel):
        drain_each_epoch: bool = True
        subcase_cache_size: int = 100
        split_across_workers: bool = True

    def __init__(
        self,
        supercase_ds: Dataset[SuperCaseType],
        supercase_loader: Callable[[SuperCaseType], List[SubCaseType]],
        cfg: Config,
        removal_decider: Optional[Callable[[CachingSubCaseDS, SubCaseType, bool], bool]] = None,
    ) -> None:
        self.supercase_ds, self.cfg, self.supercase_loader = (
            supercase_ds,
            cfg,
            supercase_loader,
        )
        self.subcases = []

        if removal_decider is None:
            self.removal_decider: Callable[
                [CachingSubCaseDS, SubCaseType, bool], bool
            ] = self.default_sampler  # type: ignore
        else:
            self.removal_decider: Callable[[CachingSubCaseDS, SubCaseType, bool], bool] = removal_decider

    @staticmethod
    # For verbosity, do not use the Python-implicit behaviour of making the current instance the first argument.
    def default_sampler(cacheds: CachingSubCaseDS, popped_case: SubCaseType, draining_phase: bool) -> bool:
        """
        By default, a case is removed whenever it is yielded
        """
        return True

    def _process_index(self, index: int, draining_phase: bool):
        subcase = self.subcases[index]
        if self.removal_decider(self, subcase, draining_phase):
            self.subcases.pop(index)
        return subcase

    def __iter__(self) -> Iterator[SubCaseType]:
        worker_info = get_worker_info()

        # First move: find out which supercases this worker should process
        if worker_info is None or not self.cfg.split_across_workers:
            supercase_indices: List[int] = list(range(len(self.supercase_ds)))  # type: ignore
            num_workers = 1  # used for cache size calculation
        else:
            num_workers = worker_info.num_workers
            worker_id = worker_info.id

            supercase_indices: List[int] = list(
                range(
                    math.ceil(len(self.supercase_ds) / num_workers) * worker_id,  # type: ignore
                    min(
                        len(self.supercase_ds),  # type: ignore
                        math.ceil(len(self.supercase_ds) / num_workers) * (worker_id + 1),  # type: ignore
                    ),
                )
            )
            logging.debug(f"Worker {worker_id} got {len(supercase_indices)} supercases.")

            if not supercase_indices:
                logging.warn(f"Worker {worker_info} had no supercases in {self}")
                return
        random.shuffle(supercase_indices)
        filling_phase = True

        for supercase_index in supercase_indices:
            supercase = self.supercase_ds[supercase_index]
            self.subcases.extend(self.supercase_loader(supercase))
            if len(self.subcases) > self.cfg.subcase_cache_size:
                filling_phase = False

            if not filling_phase:
                # Only yield samples if the cache is pretty full to increase diversity
                while len(self.subcases) > (self.cfg.subcase_cache_size // num_workers):
                    yield self._process_index(random.randint(0, len(self.subcases) - 1), draining_phase=False)

        if self.cfg.drain_each_epoch:
            # No more supercases to load, yield the remaining cases:
            # We might also skip this to keep the cache full to reduce the next epoch's startup time
            random.shuffle(self.subcases)
            while self.subcases:
                yield self._process_index(0, draining_phase=True)


class SubCaseDataset(Dataset, Generic[SubCaseType]):
    """
    Create a new dataset from a dataset which holds cases which itself hold cases.

    Common use-case: creating a 2D dataset from slices from a 3D dataset.
    In this case, fn_length_of_case might be used to determine the number of slices.
    fn_extract_case_by_index gets the slice index and returns a 2D slice.

    For determining the length of the dataset,
    the user needs to provide a function which this object applies to each supercase.
    """

    def __init__(
        self,
        src_ds: Dataset[SuperCaseType],
        fn_length_of_case: Callable[[SuperCaseType], int],
        fn_extract_case_by_index: Callable[[SuperCaseType, int], SubCaseType],
        cache_foldername: str,
        subcase_transform: Optional[Callable[[SubCaseType], SubCaseType]] = None,
    ) -> None:
        super().__init__()
        self.src_ds = src_ds
        self.fn_length_of_case = fn_length_of_case
        self.fn_extract_case_by_index = fn_extract_case_by_index
        self.cache_path = get_default_cachepath(folder_name=cache_foldername) / "sizes.pkl"
        self.transform = subcase_transform

        self.src_case_to_target_case_map: Dict[int, int] = {}
        self.first_index_of_case = {}

        if not Path(os.getenv("ML_DATA_CACHE", default="/dl_cache/")).exists():
            os.mkdir(Path(os.getenv("ML_DATA_CACHE", default="/dl_cache/")))

        if self.cache_path is not None and not get_default_cachepath(folder_name=cache_foldername).exists():
            os.mkdir(get_default_cachepath(folder_name=cache_foldername))

        if self.cache_path is not None and self.cache_path.exists():
            with open(self.cache_path, "rb") as f:
                self.src_case_to_target_case_map, self.first_index_of_case = torch.load(f)

        else:
            # Supercase stats cache needs to be recomputed
            for case_id in tqdm(range(len(cast(Sized, src_ds)))):
                case = src_ds[case_id]
                self.first_index_of_case[case_id] = len(self.src_case_to_target_case_map)
                for _ in range(fn_length_of_case(case)):
                    self.src_case_to_target_case_map[len(self.src_case_to_target_case_map)] = case_id

            if self.cache_path is not None:
                with open(self.cache_path, "wb") as f:
                    torch.save((self.src_case_to_target_case_map, self.first_index_of_case), f)

    def __len__(self) -> int:
        return len(self.src_case_to_target_case_map)

    def __getitem__(self, index: int) -> SubCaseType:
        case_id = self.src_case_to_target_case_map[index]
        case = self.src_ds[case_id]
        case_index = index - self.first_index_of_case[case_id]
        res: SubCaseType = self.fn_extract_case_by_index(case, case_index)

        if self.transform is not None:
            res = self.transform(res)

        return res


def transform_dataloader(dataloader: DataLoader, transform: Callable):
    transformed_batches = []
    for batch in dataloader:
        # The user might want to use shared blocks to process the raw data
        with torch.inference_mode():
            batch_transformed = transform(batch)
        transformed_batches.append(batch_transformed)
    return transformed_batches


def infer_stride_channels_from_features(
    features: List[torch.Tensor],
) -> Tuple[List[int], List[int]]:
    """
    Assumes the first feature map to be the raw input
    """
    channels = [v.shape[1] for v in features]
    strides = [features[0].shape[2] // v.shape[2] for v in features]
    return channels, strides


class IterableDatasetWrapper(IterableDataset):
    def __init__(self, ds: MTLDataset) -> None:
        super().__init__()
        self.ds: MTLDataset = ds

    def __iter__(self):
        """
        Yield next item from ds
        """
        if hasattr(self.ds, "_indices"):
            for idx in np.random.choice(np.arange(len(self.ds)), len(self.ds), replace=False):
                yield self.ds[idx]
        else:
            for item in self.ds:
                yield item


class CombinedDataset(IterableDataset):
    def __init__(self, mtl_datasets: List[MTLDataset]) -> None:
        super().__init__()
        self.mtl_datasets = mtl_datasets
        logging.info(f"received {len(mtl_datasets)=} and will concatenate them now")
        self.data_set = ChainDataset(self._prepare_datasets([ds for ds in self.mtl_datasets]))

    def _prepare_datasets(self, datasets):
        collection = []
        for data in datasets:
            data.batch_transform = None
            data = IterableDatasetWrapper(data)
            collection.append(data)
        return collection

    def _get_rand_ds_idx(self, num):
        return np.random.choice(np.arange(num))

    def __iter__(self):
        for item in self.data_set:
            yield item
        # tmp_ds = self._prepare_datasets([ds for ds in self.mtl_datasets])
        # while tmp_ds:
        #     idx = self._get_rand_ds_idx(len(tmp_ds))
        #     try:
        #         item = tmp_ds[idx]
        #         yield item
        #     except StopIteration:
        #         tmp_ds.pop(idx)


def combine_datasets(datasets: List[MTLDataset]) -> CombinedDataset:
    return CombinedDataset(datasets)



import os
import requests
import json


def get_nomad_allocations():
    url = os.getenv("NOMAD_ADDR") + "v1/job/" + os.getenv("NOMAD_JOB_ID") + "/allocations"
    response = requests.get(url, verify=os.getenv("NOMAD_CACERT"))
    return response.json()


def get_connection_of_rank(portname: str) -> dict[int, tuple[str, int]]:
    """
    Return a dictionary with the rank as key and a tuple with the address and the port number as value.
    """
    allocs = get_nomad_allocations()
    res: dict[int, tuple[str, int]] = {}
    for alloc_overview in allocs:
        # return alloc_overview
        if alloc_overview["ClientStatus"] != "running":
            continue
        alloc = requests.get(
            f"{os.getenv('NOMAD_ADDR')}/v1/allocation/{alloc_overview['ID']}",
            verify=os.getenv("NOMAD_CACERT"),
        ).json()
        alloc_name = alloc["Name"]
        rank = int(alloc_name[alloc_name.rfind("[") + 1 : alloc_name.rfind("]")])
        allports = alloc["Resources"]["Networks"][0]["DynamicPorts"]
        port = [p for p in allports if p["Label"] == portname][0]["Value"]
        res[rank] = alloc["Resources"]["Networks"][0]["IP"], port
    return res



from typing import Any
import numpy as np
import torch
import cv2
from shapely.affinity import scale, translate
from shapely.geometry import Polygon, shape
from rasterio.features import rasterize
from rasterio import features
import torchvision.transforms as transforms

from .GeoAnno import GeoAnno


def extract_detection_labels(
    subcase_dict: dict[str, Any],
    annos_in_window: list[GeoAnno],
    buckets: list[tuple[str, int]],
    patch_size: tuple[int, int],
):
    anno_boxes = [anno.get_enclosing_region() for anno in annos_in_window]
    anno_detection_boxes_level = [
        (
            max(0, int(anno_x)),  # / downsample_fac) - x_level),
            max(0, int(anno_y)),  # / downsample_fac) - y_level),
            min(patch_size[0] - 1, int((anno_x + anno_width))),  # / downsample_fac) - x_level),
            min(patch_size[1] - 1, int((anno_y + anno_height))),  # / downsample_fac) - y_level),
        )
        for ((anno_x, anno_y), (anno_width, anno_height)) in anno_boxes
    ]
    subcase_dict["boxes"] = torch.Tensor(anno_detection_boxes_level).float()
    subcase_dict["labels"] = torch.Tensor([x[1] for x in buckets]).long()
    return subcase_dict


def rasterize_annotations(
    window_width_level,
    window_height_level,
    annotations: list[GeoAnno],
    anno_labels: list[int],
    unlabeled_value=-1,
):
    arr = np.zeros((window_width_level, window_height_level), dtype=np.int64)
    arr.fill(unlabeled_value)
    for anno, class_key in zip(annotations, anno_labels):
        rasterize([anno.shape], out=arr, fill=unlabeled_value, default_value=(class_key))
    return arr


def rasterize_multilabel_annotations(
    window_width_level,
    window_height_level,
    annotations,
    anno_labels,
    num_classes: int,
    unlabeled_value=-1,
):
    arr = np.zeros((num_classes, window_width_level, window_height_level), dtype=np.int64)
    arr.fill(unlabeled_value)
    for anno, class_key in zip(annotations, anno_labels):
        rasterize([anno.shape], out=arr[class_key], fill=unlabeled_value, default_value=1)
    return arr


def move_anno_to_origin(anno: GeoAnno, l0_window: Polygon, downsample_fac: float):
    """
    Given an annotation detected in some window specified by its full-sized (level 0) Polygon and a downsample factor,
    returns the annotation in original l0 coordinates.
    """
    min_x, min_y, max_x, max_y = l0_window.bounds
    funcs = [
        lambda o: scale(o, xfact=downsample_fac, yfact=downsample_fac, origin=(0, 0)),  # type: ignore
        lambda o: translate(o, xoff=min_x, yoff=min_y),
    ]
    anno.shape = transforms.Compose(funcs)(anno.shape)
    return anno


def move_anno_to_window(anno: GeoAnno, l0_window: Polygon, l0_x, l0_y, downsample_fac):
    funcs = [
        lambda o: translate(o, xoff=-1 * l0_x, yoff=-1 * l0_y),
        lambda o: scale(o, xfact=1 / downsample_fac, yfact=1 / downsample_fac, origin=(0, 0)),  # type: ignore
    ]
    anno.shape = anno.shape.intersection(l0_window)
    anno.shape = transforms.Compose(funcs)(anno.shape)
    return anno


def create_featurecollection(annos: list[GeoAnno]) -> dict:
    res = {
        "type": "FeatureCollection",
        "features": [anno.to_geojson() for anno in annos],
    }
    return res


def shapes_from_binary_mask(mask: np.ndarray, downsample_fac: float, coarse: bool, min_area: float = 0.0):
    """
    Returns a list of shapely shapes from a binary mask by using connected components.

    min_area always refers to the area in level 0 (computed by your downsample_fac).
    """
    if coarse:
        fine_areas = cv2.connectedComponentsWithStats(mask, 4, cv2.CV_32S)[2][:, cv2.CC_STAT_AREA]
        mask = features.sieve(mask, size=max(1, fine_areas.mean()))
    dicts = list(features.shapes(mask))
    shapes = [(shape(x[0]), x[1]) for x in dicts]
    shapes = [
        x for x, val in shapes if x.bounds[2] - x.bounds[0] > 1.0 and x.bounds[3] - x.bounds[1] > 1.0 and val > 0.0
    ]
    if min_area:
        shapes = filter(lambda x: (x.area * (downsample_fac**2)) > min_area, shapes)
    return [scale(x, xfact=downsample_fac, yfact=downsample_fac, origin=(0, 0)) for x in shapes]  # type:ignore


def annotations_from_mask(
    mask: np.ndarray,
    for_values: dict[int, str],
    downsample_fac: float,
    coarse: bool = False,
    min_area: float = 0.0,
) -> list[GeoAnno]:
    """
    min_area refers to the area in level 0 (computed by your downsample_fac).
    """
    res: list[GeoAnno] = []
    for val, class_name in for_values.items():
        shapes_for_val = shapes_from_binary_mask(
            (mask == val).astype(np.uint8), downsample_fac, coarse, min_area=min_area
        )
        annos_for_val = [
            GeoAnno.from_shapely(shape, class_info=(class_name, len(for_values), val)) for shape in shapes_for_val
        ]
        res.extend(annos_for_val)
    return res



from typing import List, Protocol


class EncoderModel(Protocol):
    def get_feature_pyramid_channels(self) -> List[int]:
        """
        Returns the channels of the feature pyramid, starting with the channels of the input.
        """
        raise NotImplementedError

    def get_strides(self) -> List[int]:
        """
        Return the strides of the feature pyramid
        """
        raise NotImplementedError
